/** The AST node for boolean and ('&&') binary operator expressions. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class AndExpr extends BooleanBinopExpr {
    // Constructor
    public AndExpr(Expr a1, Expr a2, int ln) {
	super(a1, a2, ln);
    }


    // AST printing

    /** return the printable name of this operator */
    protected String opString() {
	return "&&";
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// && is a *short-circuiting* operator, so "a && b" really
	// evaluates like "if a then b else false"

	// evaluate the first argument
	Value arg1_value = arg1.evaluate(env);

	if (arg1_value.getBooleanValue(line_number)) {
	    // the first argument was true, so need to evaluate the
	    // second argument to determine the result of the
	    // operation
	    return arg2.evaluate(env);
	} else {
	    // we know the result will be false no matter what the
	    // second argument is, so don't evaluate the second
	    // argument (the definition of short-circuiting)
	    return new BooleanValue(false);
	}
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// && is a *short-circuiting* operator, so "a && b" really
	// evaluates like "if a then b else false"

	// get a new variable to use for the result
	ILVar result_var = 
	    il_fun_decl.newTempVar(new ResolvedBooleanType().lower());

	// construct the first argument subexpression
	ILExpr arg1_expr = arg1.lower(il_fun_decl);

	// allocate a fresh label for the true branch
	ILLabel true_label = il_fun_decl.newLabel();
	
	// generate a conditional branch for the test being true
	il_fun_decl.addStmt(new ILConditionalBranchTrueStmt(arg1_expr, 
							    true_label));

	// if we fall through to here, then the first argument is
	// false, and so the whole result is false, i.e., 0.
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(result_var), 
					     new ILIntConstantExpr(0)));

	// allocate a fresh label for the end of the && expression
	ILLabel done_label = il_fun_decl.newLabel();
	
	// generate an unconditional goto to the done label
	il_fun_decl.addStmt(new ILGotoStmt(done_label));

	// now here's where the conditional branch should branch to
	il_fun_decl.addStmt(new ILLabelStmt(true_label));

	// construct the second argument subexpression
	ILExpr arg2_expr = arg2.lower(il_fun_decl);

	// return the result of the second argument
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(result_var),
					     arg2_expr));

	// now here's where the true and the false branches merge back
	// together
	il_fun_decl.addStmt(new ILLabelStmt(done_label));

	// whew!

	// return a read of the result var
	return new ILVarExpr(result_var);
    }
}
