/** The abstract superclass of all binary operator expression AST
 * nodes that are for numeric arithmetic operations, e.g., +, -, *,
 * and /. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public abstract class ArithmeticBinopExpr extends BinopExpr {
    // Constructor
    public ArithmeticBinopExpr(Expr a1, Expr a2, int ln) {
	super(a1, a2, ln);
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result */
    protected ResolvedType typecheckHelper(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// typecheck the arguments
	ResolvedType arg1_type = arg1.typecheck(st);
	ResolvedType arg2_type = arg2.typecheck(st);

	// verify that they're ints
	arg1_type.checkIsIntType("first argument", line_number);
	arg2_type.checkIsIntType("second argument", line_number);

	// yields an int
	return new ResolvedIntType();
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the arguments
	Value arg1_value = arg1.evaluate(env);
	Value arg2_value = arg2.evaluate(env);

	// compute & return the result
	return new IntValue(
	    evaluateIntOperation(
		arg1_value.getIntValue(line_number),
		arg2_value.getIntValue(line_number)));
    }

    /** Evaluate the integer version of this operation */
    protected abstract int evaluateIntOperation(int arg1_value, 
						int arg2_value);


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// construct the argument subexpressions
	ILExpr arg1_expr = arg1.lower(il_fun_decl);
	ILExpr arg2_expr = arg2.lower(il_fun_decl);

	// return an IL expression for computing the result
	return loweredIntOperation(arg1_expr, arg2_expr);
    }

    /** Return the integer version of this operation */
    protected abstract ILExpr loweredIntOperation(ILExpr arg1_expr, 
						  ILExpr arg2_expr);
}
