/** The abstract superclass of all binary operator expression AST
 * nodes that are for equality/disequality comparison operations,
 * e.g., == and !=. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public abstract class EqualityBinopExpr extends BinopExpr {
    // Constructor
    public EqualityBinopExpr(Expr a1, Expr a2, int ln) {
	super(a1, a2, ln);
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result */
    protected ResolvedType typecheckHelper(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// typecheck the arguments
	ResolvedType arg1_type = arg1.typecheck(st);
	ResolvedType arg2_type = arg2.typecheck(st);

	// we allow this operator to be overloaded: either both
	// arguments are booleans, or both arguments are integers

	if (arg1_type.isBooleanType()) {
	    // resolved to the boolean version

	    // verify that the argument is also a boolean
	    arg2_type.checkIsBooleanType("second argument", line_number);

	} else {
	    // resolved to the integer version

	    // verify that they're integers
	    arg1_type.checkIsIntType("first argument", line_number);
	    arg2_type.checkIsIntType("second argument", line_number);
	}

	// yields a boolean
	return new ResolvedBooleanType();
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the arguments
	Value arg1_value = arg1.evaluate(env);
	Value arg2_value = arg2.evaluate(env);

	// see which version we're evaluating
	if (arg1.getResultType().isBooleanType() &&
	    arg2.getResultType().isBooleanType()) {
	    // resolved to the boolean version

	    // compute & return the result
	    return new BooleanValue(
		evaluateBooleanOperation(
		    arg1_value.getBooleanValue(line_number),
		    arg2_value.getBooleanValue(line_number)));

	} else if (arg1.getResultType().isIntType() &&
		   arg2.getResultType().isIntType()) {
	    // resolved to the int version

	    // compute & return the result
	    return new BooleanValue(
		evaluateIntOperation(
		    arg1_value.getIntValue(line_number),
		    arg2_value.getIntValue(line_number)));

	} else {
	    throw new InternalCompilerError("unexpected overload");
	}
    }

    /** Evaluate the boolean version of this operation */
    protected abstract boolean evaluateBooleanOperation(boolean arg1_value, 
							boolean arg2_value);

    /** Evaluate the integer version of this operation */
    protected abstract boolean evaluateIntOperation(int arg1_value, 
						    int arg2_value);


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// construct the argument subexpressions
	ILExpr arg1_expr = arg1.lower(il_fun_decl);
	ILExpr arg2_expr = arg2.lower(il_fun_decl);

	// see which version we're generating code for
	if (arg1.getResultType().isBooleanType() &&
	    arg2.getResultType().isBooleanType()) {
	    // resolved to the boolean version

	    // return an IL expression for computing the result
	    return loweredBooleanOperation(arg1_expr, arg2_expr);

	} else if (arg1.getResultType().isIntType() &&
		   arg2.getResultType().isIntType()) {
	    // resolved to the int version

	    // return an IL expression for computing the result
	    return loweredIntOperation(arg1_expr, arg2_expr);

	} else {
	    throw new InternalCompilerError("unexpected overload");
	}
    }

    /** Return the boolean version of this operation */
    protected abstract ILExpr loweredBooleanOperation(ILExpr arg1_expr, 
						      ILExpr arg2_expr);

    /** Return the integer version of this operation */
    protected abstract ILExpr loweredIntOperation(ILExpr arg1_expr, 
						  ILExpr arg2_expr);
}
