/** The abstract superclass of all expression AST nodes. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public abstract class Expr extends ASTNode {
    // Instance variable

    /** the result type of this expression, computed during
     * typechecking */
    protected ResolvedType result_type;


    // Constructor
    public Expr(int ln) {
	super(ln);
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * remembering the result type, returning the type of the
     * result. */
    public ResolvedType typecheck(CodeSymbolTable t) 
		throws TypecheckCompilerException {
	// typecheck this expression, remembering the result type
	result_type = typecheckHelper(t);
	// return the result type
	return result_type;
    }

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result. Implemented in subclasses. */
    protected abstract ResolvedType typecheckHelper(CodeSymbolTable t) 
	throws TypecheckCompilerException;

    /** Return the result type of this expression, after typechecking */
    public ResolvedType getResultType() {
	if (result_type == null) {
	    throw new InternalCompilerError("asking for the result type " +
					    "before it's been computed");
	}
	return result_type;
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public abstract Value evaluate(CodeEnvironment env) 
	throws EvalCompilerException;


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public abstract ILExpr lower(ILFunDecl il_fun_decl);

    /** Return the lowered result type of this expression, after
     * typechecking */
    public ILType getLoweredResultType() {
	return getResultType().lower();
    }

    /** return an augmented IL expr that converts the IL version of
     * this expression to a corresponding IL version of the target
     * type.  this is a no-op if the receiver's type is already a
     * subtype of the target type and so no conversion is
     * necessary. */
    protected ILExpr convertIfNeeded(ILExpr expr, ResolvedType type) {
	if (result_type.isSubtype(type)) {
	    // no conversion required
	    return expr;
	}
	// this needs a conversion, or there's some other error
	throw new InternalCompilerError("conversion should be implemented");
    }

    /** a helper function, used by some subclasses, to generate code
     * to check whether the argument variable is null, and throw a
     * null-pointer exception if so */
    protected void generateNullPtrCheck(ILVar var, ILFunDecl il_fun_decl) {
	// construct a comparison of var != 0 (i.e. null)
	ILExpr test_expr = new ILIntNotEqualExpr(new ILVarExpr(var), 
						 new ILIntConstantExpr(0));

	// allocate a fresh label for where to go if OK
	ILLabel ok_label = il_fun_decl.newLabel();
	
	// generate a conditional branch if true
	il_fun_decl.addStmt(new ILConditionalBranchTrueStmt(test_expr, 
							    ok_label));

	// if we fall through to here, then the ptr was null
	il_fun_decl.addStmt(
	    new ILThrowExceptionStmt("dereferencing a null pointer"));
	// this statement doesn't fall through

	// here's where we should go to if non-null
	il_fun_decl.addStmt(new ILLabelStmt(ok_label));
    }
}
