/** The AST node for an 'if' statement. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class IfStmt extends Stmt {
    // Instance variables

    // AST subtrees:
    protected Expr test;
    protected Stmt then_stmt;
    protected Stmt else_stmt;


    // Constructor
    public IfStmt(Expr t, Stmt ts, Stmt es, int ln) {
	super(ln);
	test = t;
	then_stmt = ts;
	else_stmt = es;
    }


    // AST printing
    public String toString(String indent) {
	String nested_indent = nested_indent(indent);
	String s = "if (" + test.toString(nested_indent) + ") " +
	    then_stmt.toString(indent) + 
	    "\n" + indent + "else " + else_stmt.toString(indent);
	return s;
    }


    // Typechecking operations

    /** Typecheck this statement in the given enclosing context */
    public void typecheck(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// typecheck the test expression
	ResolvedType test_type = test.typecheck(st);

	// verify that the test is a boolean
	test_type.checkIsBooleanType("test expression", line_number);

	// typecheck the then and else statements
	then_stmt.typecheck(st);
	else_stmt.typecheck(st);
    }


    // Evaluation operations

    /** Evaluate this statement in the given environment */
    public void evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the test expression
	Value test_value = test.evaluate(env);

	// see which branch to take (don't evaluate both!)
	if (test_value.getBooleanValue(line_number)) {
	    // evaluate the then stmt
	    then_stmt.evaluate(env);
	} else {
	    // evaluate the else stmt
	    else_stmt.evaluate(env);
	}
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with lowered code
     * for this statement */
    public void lower(ILFunDecl il_fun_decl) {
	// construct the test subexpression
	ILExpr test_expr = test.lower(il_fun_decl);

	// allocate a fresh label for the false branch
	ILLabel false_label = il_fun_decl.newLabel();
	
	// generate a conditional branch for the test being false
	il_fun_decl.addStmt(new ILConditionalBranchFalseStmt(test_expr, 
							     false_label));

	// if we fall through to here, then we're in the true branch

	// generate the code for the true stmts
	then_stmt.lower(il_fun_decl);

	// we have to branch over the else code

	// allocate a fresh label for the end of the if stmt
	ILLabel done_label = il_fun_decl.newLabel();

	// generate an unconditional goto to the done label
	il_fun_decl.addStmt(new ILGotoStmt(done_label));

	// now here's where the conditional branch should branch to
	il_fun_decl.addStmt(new ILLabelStmt(false_label));

	// generate the code for the else stmts
	else_stmt.lower(il_fun_decl);

	// now here's where the true and the false branches merge back
	// together
	il_fun_decl.addStmt(new ILLabelStmt(done_label));

	// whew!
    }
}
