/** The AST node for the special main class declaration. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class MainClassDecl extends ClassDecl {
    // Instance variables

    // AST subtrees:
    // class_name declared in superclass
    protected String main_method_argv_name;
    protected List<Stmt> main_method_body;

    // symbol_table declared in superclass


    // Constructor
    public MainClassDecl(String cn, String mman, List<Stmt> mmb, int ln) {
	super(cn, ln);
	main_method_argv_name = mman;
	main_method_body = mmb;
    }


    // AST printing
    public String toString(String indent) {
	String nested_indent = nested_indent(indent);
	String nested_nested_indent = nested_indent(nested_indent);
	return
	    "class " + class_name + " {\n" +
	    nested_indent + "public static void main(String[] " + 
	    main_method_argv_name + ") {" +
	    prefixedListToString(main_method_body, nested_nested_indent,
				 "\n" + nested_nested_indent) +
	    nested_indent + "}\n" +
	    indent + "}";
    }


    // Typechecking operations

    /** A helper that just declares this class decl */
    public void declareClass(GlobalSymbolTable st)
		throws TypecheckCompilerException {
	// create an empty symbol table for this class.  (it will get
	// filled in by later passes.)
	symbol_table = new ClassSymbolTable(class_name, st);
	
	// record this new class in the global scope
	st.declareClass(class_name, symbol_table, line_number);
    }

    /** A helper that checks and fills in the class's interface, given
     * that all classes in the program have been declared. */
    public void computeClassInterface() throws TypecheckCompilerException {
	// the main class has neither class variables nor
	// (non-static) methods, so the interface remains empty
    }

    /** A helper that finishes typechecking of the body of the class
     * declaration, once the full class interfaces of all the
     * program's classes have been constructed. */
    public void typecheckClass() throws TypecheckCompilerException {
	// create a SymbolTable for the main method's body
	CodeSymbolTable method_symbol_table = 
	    new CodeSymbolTable(symbol_table);
	
	// typecheck the main method's body statements in this symbol table
	Iterator<Stmt> iter = main_method_body.iterator();
	while (iter.hasNext()) {
	    Stmt stmt = iter.next();
	    stmt.typecheck(method_symbol_table);
	}
    }


    // Evaluation operations

    /** Evaluate the declaration part of the class. */
    public void evaluateDecl(GlobalEnvironment global_env)
		throws EvalCompilerException {
	// create the class's environment
	ClassEnvironment class_env = 
	    new ClassEnvironment(class_name, global_env);
       
	// (no superclass, class variables, or non-static methods to
	// declare in this class)

	// register the class's environment in the enclosing, global
	// environment
	global_env.declareClass(class_name, class_env, line_number);
    }	

    /** Evaluate the main method of this class, throwing a
     * CompilerException if there is an error during evaluation. */
    public void evaluateMain(Environment global_env, boolean printCalls)
		throws EvalCompilerException {
	// look up the main class's environment
	ClassEnvironment class_env = 
	    global_env.lookupClass(class_name, line_number);

	// create an environment for the main method's body inside the
	// class environment
	MethodEnvironment method_env = new MethodEnvironment(class_env, null);

	// evaluate the main method's body statements in its environment
	Iterator<Stmt> iter = main_method_body.iterator();
	while (iter.hasNext()) {
	    Stmt stmt = iter.next();
	    stmt.evaluate(method_env);
	}
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language program with declarations for
     * this class */
    public void lower(ILProgram il_program) {
	// generate a main function for the program

	// create the lowered function declaration
	ILType int_type = new ResolvedIntType().lower();
	ILFunDecl il_fun_decl = new ILFunDecl(il_program, int_type, "main");
	// no formal parameters for this function

	// generate code for the statements
	Iterator<Stmt> iter = main_method_body.iterator();
	while (iter.hasNext()) {
	    Stmt stmt = iter.next();
	    stmt.lower(il_fun_decl);
	}

	// return 0 (which means success)
	il_fun_decl.addStmt(new ILReturnStmt(new ILIntConstantExpr(0)));

	// finally, add the main function to the program
	il_program.addFunDecl(il_fun_decl);

	// declare the class's (empty) vtbl in initialized global data
	declareVtbl(il_program);
    }
}
