/** The AST node for a MethodDecl */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class MethodDecl extends ASTNode {
    // Instance variables
    protected Type return_type;
    protected String name;
    protected List<VarDecl> formals;
    protected List<Stmt> body;
    protected Expr return_expr;

    // The information for the formals, computing during typechecking
    protected List<LocalVarInterface> formal_interfaces;


    // Constructor
    public MethodDecl(Type rt, String n, List<VarDecl> fs, 
		      List<Stmt> b, Expr re, int ln) {
	super(ln);
	return_type = rt;
	name = n;
	formals = fs;
	body = b;
	return_expr = re;
    }


    // AST printing
    public String toString(String indent) {
	String nested_indent = nested_indent(indent);
	return return_type.toString(indent) + " " + name + "(" + 
	    separatedListToString(formals, nested_indent, ", ") + ") {" + 
	    prefixedListToString(body, nested_indent, "\n" + nested_indent) +
	    "\n" + nested_indent + 
	    "return " + return_expr.toString(nested_indent) + ";\n" +
	    indent + "}";
    }


    // Typechecking operations

    /** A helper that computes the interface to a method declaration,
     * for storing in a class's symbol table. */
    public MethodInterface computeMethodInterface(SymbolTable st) 
		throws TypecheckCompilerException {
	// compute the method's declared return type
	ResolvedType resolved_return_type = return_type.resolve(st);

	// compute the method's declared formal types
	List<ResolvedType> formal_types = new LinkedList<ResolvedType>();
	Iterator<VarDecl> iter = formals.iterator();
	while (iter.hasNext()) {
	    VarDecl formal = iter.next();
	    formal_types.add(formal.type.resolve(st));
	}

	return new MethodInterface(resolved_return_type, formal_types);
    }

    /** A helper that finishes typechecking of the body of the method
     * declaration, once the full class interfaces of all the
     * program's classes have been constructed. */
    public void typecheckMethod(ResolvedType receiver_type, 
				SymbolTable enclosing_st)
		throws TypecheckCompilerException {
	// create a symbol table in which to typecheck this method's body
	CodeSymbolTable symbol_table = new CodeSymbolTable(enclosing_st);

	formal_interfaces = new LinkedList<LocalVarInterface>();

	// declare the "this" implicit variable
	formal_interfaces.add(
	    symbol_table.declareLocalVariable(receiver_type, "this", 
					      line_number));

	// declare each of the method's formals
	Iterator<VarDecl> formals_iter = formals.iterator();
	while (formals_iter.hasNext()) {
	    VarDecl formal = formals_iter.next();
	    formal_interfaces.add(
	        symbol_table.declareLocalVariable(formal.type.resolvedType(), 
						  formal.name, line_number));
	}

	// typecheck the body statements
	Iterator<Stmt> body_iter = body.iterator();
	while (body_iter.hasNext()) {
	    Stmt stmt = body_iter.next();
	    stmt.typecheck(symbol_table);
	}

	// typecheck the result expression
	ResolvedType return_expr_type = return_expr.typecheck(symbol_table);

	// verify that the returned expression can be assigned to the
	// method's declared result type
	ResolvedType declared_return_type = return_type.resolvedType();
	return_expr_type.checkIsAssignableTo(declared_return_type,
					     "returned expression",
					     "declared method result type", 
					     line_number);
    }


    // Evaluation operations

    /** Evaluate this method, declared in the given class, when called
     * with the given receiver and arguments, return the value
     * returned by the method. */
    public Value evaluate(Value receiver, List<Value> args, 
			  ClassEnvironment enclosing_env,
			  CodeEnvironment calling_env)
		throws EvalCompilerException {
	if (enclosing_env.printCalls()) {
	    String s = 
		calling_env.caller_depth_indent() + 
		"calling " + receiver.toString() + "." + name + "(";
	    Iterator<Value> iter = args.iterator();
	    boolean first = true;
	    while (iter.hasNext()) {
		Value arg = iter.next();
		if (first) {
		    first = false;
		} else {
		    s += ", ";
		}
		s += arg.toString();
	    }
	    s += ")...";
	    System.out.println(s);
	}

	// create an environment for the method's body
	MethodEnvironment method_env = new MethodEnvironment(enclosing_env,
							     calling_env);

	// declare the implicit "this" formal value and initialize it
	// with the receiver object
	method_env.declareLocalVariable("this", receiver, line_number);

	// declare and initialize each of the method's formals with
	// the corresponding actual arguments
	Iterator<VarDecl> formals_iter = formals.iterator();
	Iterator<Value> args_iter = args.iterator();
	while (formals_iter.hasNext()) {
	    VarDecl formal = formals_iter.next();
	    Value arg = args_iter.next();
	    method_env.declareLocalVariable(formal.name, arg, line_number);
	}

	// evaluate the body statements in this environment
	Iterator<Stmt> iter = body.iterator();
	while (iter.hasNext()) {
	    Stmt stmt = iter.next();
	    stmt.evaluate(method_env);
	}

	// evaluate the returned expression
	Value result_value = return_expr.evaluate(method_env);

	// convert the returned value to the declared return type, if needed
	result_value = 
	    result_value.convertValueIfNeeded(return_type.resolvedType(), 
					      line_number);

	if (enclosing_env.printCalls()) {
	    String s = 
		calling_env.caller_depth_indent() + 
		"returning " + result_value.toString() + " from " + name;
	    System.out.println(s);
	}

	// return the method's result value
	return result_value;
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language program with a global
     * function declaration for this method, where the symbol table is
     * for the enclosing class. */
    public void lower(ILProgram il_program, ClassSymbolTable class_st) {
	// compute the lowered return type
	ILType il_return_type = return_type.lower();

	// compute the lowered function name by concatenating the name
	// of the enclosing class with the name of the method
	String il_fun_name = class_st.getClassName() + "." + name;

	// create the lowered function declaration
	ILFunDecl il_fun_decl = 
	    new ILFunDecl(il_program, il_return_type, il_fun_name);

	// declare all the formals (including the implicit "this" formal)
	Iterator<LocalVarInterface> formals_iter = 
	    formal_interfaces.iterator();
	while (formals_iter.hasNext()) {
	    LocalVarInterface formal_interface = formals_iter.next();
	    il_fun_decl.declareFormal(formal_interface);
	}

	// generate code for the statements
	Iterator<Stmt> body_iter = body.iterator();
	while (body_iter.hasNext()) {
	    Stmt stmt = body_iter.next();
	    stmt.lower(il_fun_decl);
	}

	// generate code for the result expression
	ILExpr result_expr = return_expr.lower(il_fun_decl);

	// convert the returned value to the declared return type, if needed
	result_expr = return_expr.convertIfNeeded(result_expr, 
						  return_type.resolvedType());

	// generate the return statement
	il_fun_decl.addStmt(new ILReturnStmt(result_expr));

	// finally, add the function to the program
	il_program.addFunDecl(il_fun_decl);
    }
}
