/** The AST node for a new expression. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class NewExpr extends Expr {
    // Instance variables

    // AST subtrees:
    protected String class_name;


    // Constructor
    public NewExpr(String cn, int ln) {
	super(ln);
	class_name = cn;
    }


    // AST printing
    public String toString(String indent) {
	return "new " + class_name + "()";
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result */
    protected ResolvedType typecheckHelper(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// yields the type of the class being instantiated
	return new ResolvedClassType(class_name, st, line_number);
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// look up the environment for the class being instantiated
	ClassEnvironment class_env = env.lookupClass(class_name, line_number);

	// create the new instance value
	ClassValue instance = new ClassValue(class_env);

	// initialize all its instance variables
	ClassSymbolTable ci = getResultType().getClassInterface();
	ci.initializeInstanceVariables(instance);

	// return the new instance
	return instance;
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// compute the lowered type of the result
	ILType lowered_result_type = getLoweredResultType();

	// compute the lowered type of the record being allocated
	ILType record_type = lowered_result_type.getPtrTargetType();

	// generate an object allocation instruction
	ILExpr result_expr = new ILAllocateExpr(record_type);

	// store it into a variable, so we can refer to it more than once
	ILVar result_var = il_fun_decl.newTempVar(lowered_result_type);
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(result_var),
					     result_expr));

	// initialize the vtbl member of the new instance

	// the name of the global variable which is this class's vtbl record:
        ClassSymbolTable class_st = getResultType().getClassInterface();
	String vtbl_name = class_st.getClassName() + ".vtbl";
	ILGlobalVar vtbl_var = il_fun_decl.lookupGlobalVar(vtbl_name);

	// take the address of this global variable
	ILExpr vtbl_addr_expr = new ILGlobalAddressExpr(vtbl_var);

	// get the vtbl record member
	ILRecordMember vtbl_member = record_type.getRecordMember("<vtbl>");

	// store this address into the new object's vtbl field
	ILAssignableExpr vtbl_field =
	    new ILFieldAccessExpr(new ILVarExpr(result_var), 
				  record_type, vtbl_member);
	il_fun_decl.addStmt(new ILAssignStmt(vtbl_field, vtbl_addr_expr));

	// initialize all the instance variables of the newly
	// allocated object to their default value
	class_st.generateInstanceVariableInits(result_var, il_fun_decl);

	// return the result
	return new ILVarExpr(result_var);
    }
}
