/** The AST node for a Program */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class Program extends ASTNode {
    // Instance variables

    // AST subtrees:
    protected MainClassDecl main_class_decl;
    protected List<RegularClassDecl> class_decls;


    // Constructor
    public Program(MainClassDecl mcd, List<RegularClassDecl> cds, int ln) {
	super(ln);
	main_class_decl = mcd;
	class_decls = cds;
    }


    // AST printing
    public String toString(String indent) {
	return 
	    main_class_decl.toString(indent) +
	    prefixedListToString(class_decls, indent, "\n" + indent);
    }


    // Typechecking operations

    /** Typecheck a whole program, throwing a TypecheckCompilerException if the
     * program doesn't typecheck */
    public GlobalSymbolTable typecheck() throws TypecheckCompilerException {
	// typechecking a program is done in several phases, since
	// classes, methods, etc. later in the program can legally be
	// referenced earlier in the program.

	// the results of all the passes other than the final pass are
	// accumulated in the symbol table for the global scope.
	GlobalSymbolTable global_symbol_table = new GlobalSymbolTable();

	// first, we need to collect the names of all the classes in
	// the program.  (we will use this information to identify
	// which class types are legal.)
	declareClasses(global_symbol_table);

	// second, we compute the interface of each class.  (we will
	// use this information to check references to a class,
	// e.g. method calls.)
	computeClassInterfaces();

	// finally, we do the remaining typechecking of each class in
	// the program, including checking all its methods and their
	// bodies.
	typecheckClasses();

	// in case the client wants it, return the final global sym table
	return global_symbol_table;
    }

    /** A helper that creates empty class symbol tables for each class
     * declared in the program */
    protected void declareClasses(GlobalSymbolTable st)
		throws TypecheckCompilerException {
	main_class_decl.declareClass(st);
	Iterator<RegularClassDecl> iter = class_decls.iterator();
	while (iter.hasNext()) {
	    RegularClassDecl cd = iter.next();
	    cd.declareClass(st);
	}
    }

    /** A helper that checks and fills in the interface to each class
     * declared in the program */
    protected void computeClassInterfaces() throws TypecheckCompilerException {
	main_class_decl.computeClassInterface();
	Iterator<RegularClassDecl> iter = class_decls.iterator();
	while (iter.hasNext()) {
	    RegularClassDecl cd = iter.next();
	    cd.computeClassInterface();
	}
    }

    /** A helper that finishes typechecking of the body of each class
     * declaration, once the full class interfaces have been
     * constructed. */
    protected void typecheckClasses() throws TypecheckCompilerException {
	main_class_decl.typecheckClass();
	Iterator<RegularClassDecl> iter = class_decls.iterator();
	while (iter.hasNext()) {
	    RegularClassDecl cd = iter.next();
	    cd.typecheckClass();
	}
    }


    // Evaluation operations

    /** Evaluate a whole program, throwing an EvalCompilerException if
     * there is an error during evaluation. */
    public void evaluate(boolean printCalls) throws EvalCompilerException {
	// as with typechecking, evaluation is done in several phases.

	// the results of all the passes other than the final pass are
	// accumulated in the environment for the global scope.
	GlobalEnvironment global_env = new GlobalEnvironment(printCalls);

	// first, compute the environments for all the classes in the
	// program.
	main_class_decl.evaluateDecl(global_env);
	Iterator<RegularClassDecl> iter = class_decls.iterator();
	while (iter.hasNext()) {
	    RegularClassDecl cd = iter.next();
	    cd.evaluateDecl(global_env);
	}

	// finally, we evaluate the main method of the main class
	main_class_decl.evaluateMain(global_env, printCalls);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Translate a whole program into a corresponding
     * intermediate-language program */
    public ILProgram lower() {
	// create an ILProgram object
	ILProgram il_program = new ILProgram();

	// lower each class
	main_class_decl.lower(il_program);
	Iterator<RegularClassDecl> iter = class_decls.iterator();
	while (iter.hasNext()) {
	    RegularClassDecl cd = iter.next();
	    cd.lower(il_program);
	}

	// return the lowered program
	return il_program;
    }
}
