/** The AST node for a regular class declaration, given after the main
 * class declaration. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class RegularClassDecl extends ClassDecl {
    // Instance variables

    // AST subtrees:
    // class_name declared in superclass
    protected String superclass_name; // null if there was no extends clause
    protected List<ClassVarDecl> class_variable_decls;
    protected List<MethodDecl> method_decls;

    // symbol_table declared in superclass


    // Constructor
    public RegularClassDecl(String cn, String sn, 
			    List<ClassVarDecl> cvds,
			    List<MethodDecl> mds, 
			    int ln) {
	super(cn, ln);
	superclass_name = sn;
	class_variable_decls = cvds;
	method_decls = mds;
    }


    // AST printing

    public String toString(String indent) {
	String nested_indent = nested_indent(indent);
	String nested_nested_indent = nested_indent(nested_indent);
	return
	    "class " + class_name + 
	    (superclass_name == null ? "" : " extends " + superclass_name) +
	    " {" +
	    prefixedAndTerminatedListToString(class_variable_decls, 
					      nested_indent,
					      "\n" + nested_indent, ";") +
	    prefixedListToString(method_decls, nested_indent, 
				 "\n" + nested_indent) +
	    "\n" + indent + "}";
    }


    // Typechecking operations

    /** A helper that just declares this class decl */
    public void declareClass(GlobalSymbolTable st)
		throws TypecheckCompilerException {
	// create an empty symbol table for this class.  (it will get
	// filled in by later passes.)
	symbol_table = new ClassSymbolTable(class_name, st);
	
	if (superclass_name != null) {
	    // check that the superclass names a previously declared
	    // class.  (we check this during the first pass to ensure
	    // that superclasses all are declared before their
	    // subclasses, which guarantees that there can be no
	    // cycles in the inheritance graph.)
	    ClassSymbolTable superclass_st = 
		st.lookupClass(superclass_name, line_number);

	    // remember the superclass
	    symbol_table.setSuperclass(superclass_st);
	}

	// record this new class in the global scope
	st.declareClass(class_name, symbol_table, line_number);
    }

    /** A helper that checks and fills in the class's interface, given
     * that all classes in the program have been declared. */
    public void computeClassInterface() throws TypecheckCompilerException {
	// check the class variable declarations, and add them to
	// the class's interface (as represented by its symbol table)
	Iterator<ClassVarDecl> class_vars_iter =
	    class_variable_decls.iterator();
	while (class_vars_iter.hasNext()) {
	    ClassVarDecl cvd = class_vars_iter.next();
	    cvd.computeInterface(symbol_table);
	}

	// compute the interface to the methods in this class, and add
	// them to the class's interface (as represented by its symbol
	// table)
	Iterator<MethodDecl> method_decls_iter = method_decls.iterator();
	while (method_decls_iter.hasNext()) {
	    MethodDecl md = method_decls_iter.next();

	    // compute the interface to this method
	    MethodInterface mi = md.computeMethodInterface(symbol_table);

	    // record this method interface in the class's symbol
	    // table (checking for duplicates in the process)
	    symbol_table.declareMethod(md.name, mi, md.line_number);
	}
    }

    /** A helper that finishes typechecking of the body of the class
     * declaration, once the full class interfaces of all the
     * program's classes have been constructed. */
    public void typecheckClass() throws TypecheckCompilerException {
	// compute this class's type, which will be the type of the
	// implicit receiver of its methods
	ResolvedType receiver_type = symbol_table.getType();

	// check the method declarations' bodies
	Iterator<MethodDecl> iter = method_decls.iterator();
	while (iter.hasNext()) {
	    MethodDecl md = iter.next();
	    md.typecheckMethod(receiver_type, symbol_table);
	}
    }


    // Evaluation operations

    /** Evaluate the declaration part of the class. */
    public void evaluateDecl(GlobalEnvironment global_env)
		throws EvalCompilerException {
	// create the class's environment
	ClassEnvironment class_env = 
	    new ClassEnvironment(class_name, global_env);
	
	if (superclass_name != null) {
	    // get the environment for the superclass
	    ClassEnvironment superclass_env = 
		global_env.lookupClass(superclass_name, line_number);

	    // remember the superclass's environment
	    class_env.setSuperclassEnvironment(superclass_env);
	}
       
	// register the class variable declarations
	Iterator<ClassVarDecl> class_vars_iter =
	    class_variable_decls.iterator();
	while (class_vars_iter.hasNext()) {
	    ClassVarDecl cvd = class_vars_iter.next();
	    cvd.declareVariable(class_env);
	}
       
	// register the method declarations
	Iterator<MethodDecl> method_decls_iter = method_decls.iterator();
	while (method_decls_iter.hasNext()) {
	    MethodDecl md = method_decls_iter.next();
	    class_env.declareMethod(md.name, md, md.line_number);
	}

	// register the class's environment in the enclosing, global
	// environment
	global_env.declareClass(class_name, class_env, line_number);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language program with declarations for
     * this class */
    public void lower(ILProgram il_program) {
	// lower the class variable declarations into whatever is
	// needed
	Iterator<ClassVarDecl> class_vars_iter =
	    class_variable_decls.iterator();
	while (class_vars_iter.hasNext()) {
	    ClassVarDecl cvd = class_vars_iter.next();
	    cvd.lower(il_program, symbol_table);
	}
       
	// lower each of the method declarations into global function
	// declarations
	Iterator<MethodDecl> method_decls_iter = method_decls.iterator();
	while (method_decls_iter.hasNext()) {
	    MethodDecl md = method_decls_iter.next();
	    md.lower(il_program, symbol_table);
	}

	// declare the class's vtbl in initialized global data
	declareVtbl(il_program);
    }
}
