/** The AST node for an initialized local var decl statement. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class VarDeclStmt extends Stmt {
    // Instance variables

    // AST subtrees:
    protected VarDecl decl;  // the variable name and type
    protected Expr rhs;      // the initializing expression

    // The information about the declared variable, set after typechecking
    protected LocalVarInterface var_interface;


    // Constructor
    public VarDeclStmt(VarDecl d, Expr r, int ln) {
	super(ln);
	decl = d;
	rhs = r;
    }


    // AST printing
    public String toString(String indent) {
	return decl.toString(indent) + " = " + rhs.toString(indent) + ";";
    }


    // Typechecking operations

    /** Typecheck this statement in the given enclosing context */
    public void typecheck(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// compute the type of the declared variable
	ResolvedType lhs_type = decl.type.resolve(st);

	// typecheck the rhs expression
	ResolvedType rhs_type = rhs.typecheck(st);

	// verify that the type of the rhs can be assigned to the type
	// of the lhs
	rhs_type.checkIsAssignableTo(lhs_type, "rhs", "lhs", line_number);

	// finally, declare the new local variable, checking for duplicates
	var_interface = 
	    st.declareLocalVariable(lhs_type, decl.name, line_number);
    }


    // Evaluation operations

    /** Evaluate this statement in the given environment */
    public void evaluate(CodeEnvironment env) throws EvalCompilerException {
	// evaluate the rhs expression
	Value rhs_value = rhs.evaluate(env);

	// convert the rhs value to the type of the lhs, if needed
	ResolvedType lhs_type = decl.type.resolvedType();
	rhs_value = rhs_value.convertValueIfNeeded(lhs_type, line_number);

	// declare the variable in the current environment,
	// initialized to the rhs value
	env.declareLocalVariable(decl.name, rhs_value, line_number);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with lowered code
     * for this statement */
    public void lower(ILFunDecl il_fun_decl) {
	// get the variable's type
	ResolvedType lhs_type = decl.type.resolvedType();

	// declare this as a new local variable in the IL function
	il_fun_decl.declareLocal(var_interface);

	// construct the rhs subexpression
	ILExpr rhs_expr = rhs.lower(il_fun_decl);

	// convert the rhs to the type of the lhs, if needed
	rhs_expr = rhs.convertIfNeeded(rhs_expr, lhs_type);

	// get the IL var for this local variable
	ILVar lhs_var = il_fun_decl.lookupVar(var_interface);

	// generate the initializing assignment
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(lhs_var), 
					     rhs_expr));
    }
}
