/** The AST node for a variable reference expression.  It is also
 * assignable, i.e., it can also appear on the left-hand-side of an
 * assignment statement. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class VarExpr extends AssignableExpr {
    // Instance variables

    // AST subtrees:
    protected String name;

    // The information about the variable, set after typechecking
    protected VarInterface var_interface;


    // Constructor
    public VarExpr(String n, int ln) {
	super(ln);
	name = n;
    }


    // AST printing
    public String toString(String indent) {
	return name;
    }


    // Typechecking operations

    /** Typecheck this expression in the given enclosing context,
     * returning the type of the result */
    protected ResolvedType typecheckHelper(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// look up and remember the information about the variable
	var_interface = st.lookupVariable(name, line_number);
	
	// return the type of the variable
	return var_interface.getType();
    }


    // Evaluation operations

    /** Evaluate this expression in the given environment, returning
     * the value of the result */
    public Value evaluate(CodeEnvironment env) throws EvalCompilerException {
	// look up the variable's current value
	return var_interface.lookupVariable(env, line_number);
    }

    /** Evaluate an assignment to this expression of the given value
     * in the given environment */
    public void evaluateAssignment(CodeEnvironment env, Expr rhs) 
	    throws EvalCompilerException {
	// evaluate the rhs expression
	Value rhs_value = rhs.evaluate(env);

	// convert the rhs value to the type of the lhs, if needed
	ResolvedType lhs_type = getResultType();
	rhs_value = rhs_value.convertValueIfNeeded(lhs_type, line_number);

	// update the variable's current value
	var_interface.assignVariable(env, rhs_value, line_number);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with code to
     * evaluate this expression, returning an IL expression for
     * computing the result */
    public ILExpr lower(ILFunDecl il_fun_decl) {
	// generate the variable read, based on the kind of variable
	// (class or local) we're reading
	return var_interface.generateRead(il_fun_decl);
    }

    /** Augment an intermediate-language function with lowered code to
     * assign to this expression the given rhs expression */
    public void lowerAssignment(ILFunDecl il_fun_decl, Expr rhs) {
	// construct the rhs subexpression
	ILExpr rhs_expr = rhs.lower(il_fun_decl);

	// convert the rhs to the type of the lhs, if needed
	ResolvedType lhs_type = getResultType();
	rhs_expr = rhs.convertIfNeeded(rhs_expr, lhs_type);

	// generate the assignment, based on the kind of variable
	// (class or local) we're assigning to
	var_interface.generateAssignment(rhs_expr, il_fun_decl);
    }
}
