/** The AST node for a 'while' statement. */

package AST;

import Typechecker.*;
import Evaluator.*;
import IL.*;
import Helpers.*;

public class WhileStmt extends Stmt {
    // Instance variables

    // AST subtrees:
    protected Expr test;
    protected Stmt body;


    // Constructor
    public WhileStmt(Expr t, Stmt b, int ln) {
	super(ln);
	test = t;
	body = b;
    }


    // AST printing
    public String toString(String indent) {
	String nested_indent = nested_indent(indent);
	return
	    "while (" + test.toString(nested_indent) + ") " + 
	    body.toString(indent);
    }


    // Typechecking operations

    /** Typecheck this statement in the given enclosing context */
    public void typecheck(CodeSymbolTable st)
		throws TypecheckCompilerException {
	// typecheck the test expression
	ResolvedType test_type = test.typecheck(st);

	// verify that the test is a boolean
	test_type.checkIsBooleanType("test expression", line_number);

	// typecheck the body statement (which is always in a loop)
	body.typecheck(st);
    }


    // Evaluation operations

    /** Evaluate this statement in the given environment */
    public void evaluate(CodeEnvironment env) throws EvalCompilerException {
	while (true) {
	    // evaluate the test expression
	    Value test_value = test.evaluate(env);

	    // see if the while body should be evaluated
	    if (test_value.getBooleanValue(line_number)) {
		// evaluate the body stmt
		body.evaluate(env);
		// loop back to the top
	    } else {
		// done with the while loop
		break;
	    }
	}
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** Augment an intermediate-language function with lowered code
     * for this statement */
    public void lower(ILFunDecl il_fun_decl) {
	// allocate a fresh label for the top of the loop
	ILLabel start_label = il_fun_decl.newLabel();
	
	// generate the label at the top of the loop
	il_fun_decl.addStmt(new ILLabelStmt(start_label));

	// construct the test subexpression
	ILExpr test_expr = test.lower(il_fun_decl);

	// allocate a fresh label for the end of the while stmt
	ILLabel done_label = il_fun_decl.newLabel();
	
	// generate a conditional branch for the test being false
	// (including the code for the test inside the loop, since we
	// want to reevaluate it each time through the loop)
	il_fun_decl.addStmt(new ILConditionalBranchFalseStmt(test_expr, 
							     done_label));

	// if we fall through to here, then the test was true, and we
	// should do the loop body

	// generate the loop body
	body.lower(il_fun_decl);

	// now branch back to the top of the loop, and start all over
	il_fun_decl.addStmt(new ILGotoStmt(start_label));

	// here's where the loop-exit branch should go to
	il_fun_decl.addStmt(new ILLabelStmt(done_label));

	// whew!
    }
}
