/** This is the abstract superclass that serves as the interface to
 * the capabilities of the target machine for target code generation,
 * from the point of view of the (IL) intermediate language.  Concrete
 * subclasses implement specific machines. */

package CodeGen;

import IL.*;
import java.io.*;
import java.util.*;

public abstract class Target {
    /** whether or not we should print out each emitted line of code
     * as it's generated to the output file */
    protected boolean printCode;

    /** whether or not we should add an implicit underscore as a
     * prefix to the names of functions in the runtime library */
    protected boolean runtimeFunsUnderscorePrefix;

    /** the output file, set by initialize */
    protected Writer output_file;


    /** Constructor */
    public Target() { }

    /** Enable optimizations (disabled by default) */
    public abstract void enableOptimizations();


    // Initialization operations

    /** Initialize a the target code generator */
    public void initialize(String input_file_name, boolean printCode,
			   boolean runtimeFunsUnderscorePrefix) 
		throws IOException {
	// remember control flags
	this.printCode = printCode;
	this.runtimeFunsUnderscorePrefix = runtimeFunsUnderscorePrefix;
	// compute the name of the output file from the input file
	String output_file_name = 
	    removeJavaSuffix(input_file_name) + "." + outputSuffix();
	// make a buffered output writer onto this file name
	output_file = new BufferedWriter(new FileWriter(output_file_name));
    }

    /** helper to remove the ".java" suffix from the argument string,
     * if it's present */
    protected String removeJavaSuffix(String name) {
	if (name.endsWith(".java")) {
	    // remove the suffix (which is 5 chars long)
	    return name.substring(0, name.length() - 5);
	} else {
	    // no suffix to remove
	    return name;
	}
    }

    /** helper to return the suffix to use for this output file */
    protected String outputSuffix() {
	// by default, use the standard assembly language suffix
	return "s";
    }


    // Finalization operations

    /** Finish up generation of the output file */
    public void finalize() throws IOException {
	// close the output file
	output_file.close();
    }


    // Code generation operations for program beginnings and endings

    /** Generate whatever code is needed by this target at the start
     * of the output file */
    public abstract void emitProgramPrologue() throws IOException;

    /** Generate whatever code is needed by this target at the end
     * of the output file */
    public abstract void emitProgramEpilogue() throws IOException;


    // Code generation operations for function beginnings and endings

    /** Generate code to declare and introduce a function, laying out
     * the function's stack frame in the process */
    public abstract void emitFunctionPrologue(
					ILType return_type,
					String fun_name,
					List<ILFormalVarDecl> formals,
					Map<String,ILVarDecl> decls)
	throws IOException;

    /** Generate code to finish a function */
    public abstract void emitFunctionEpilogue(String fun_name)
	throws IOException;


    // Code generation operations for global declarations

    /** Generate code to declare (but not initialize) a global
     * variable */
    public abstract void emitUninitializedGlobalVarDecl(ILType type, 
							String name)
	throws IOException;

    /** Generate code to declare and initialize a global variable */
    public abstract void emitInitializedGlobalVarDecl(ILType type, 
						      String name,
						      ILValue value)
	throws IOException;

    /** Generate code to initialize a global memory location to a
     * double */
    public abstract void emitAlignDirective(ILType type) throws IOException;

    /** Generate code to initialize a global memory location to an
     * int */
    public abstract void emitIntInitializer(int value) throws IOException;

    /** Generate code to initialize a global memory location to a
     * double */
    public abstract void emitDoubleInitializer(double value)
	throws IOException;

    /** Generate code to initialize a global memory location to the
     * address of the global variable with the given name */
    public abstract void emitGlobalAddressInitializer(String name)
	throws IOException;

    /** Generate code to initialize a global memory location to the
     * address of a label */
    public abstract void emitLabelAddressInitializer(ILLabel label)
	throws IOException;


    // Code generation operations for each of the different kinds of
    // statements in the IL

    /** Generate code for an assignment */
    public abstract void emitAssign(ILAssignableExpr lhs, ILExpr rhs)
	throws IOException;

    /** Generate code for a false (zero) conditional branch */
    public abstract void emitConditionalBranchFalse(ILExpr test, 
						    ILLabel target_label)
	throws IOException;

    /** Generate code for a true (non-zero) conditional branch */
    public abstract void emitConditionalBranchTrue(ILExpr test,
						   ILLabel target_label)
	throws IOException;

    /** Generate code for an expression, ignoring the result */
    public abstract void emitExpr(ILExpr expr) throws IOException;

    /** Generate code for a goto statement */
    public abstract void emitGoto(ILLabel label) throws IOException;

    /** Generate code for a label */
    public abstract void emitLabel(ILLabel label) throws IOException;

    /** Generate code for a return */
    public abstract void emitReturn(ILExpr result) throws IOException;

    /** Generate code to signal an error */
    public abstract void emitThrowException(String message) throws IOException;


    // Code generation operations for each of the different kinds of
    // expressions in the IL

    /** Generate code to read the contents of a local variable */
    public abstract Location emitVarRead(ILVarDecl decl) throws IOException;

    /** Generate code to assign to a local variable */
    public abstract void emitVarAssign(ILVarDecl decl, Location rhs_location)
	throws IOException;

    /** Generate code to read the target of a pointer that points to a
     * value of the given type */
    public abstract Location emitPtrRead(ILExpr ptr, ILType type) 
	throws IOException;

    /** Generate code to assign to the target of a pointer that points
     * to a value of the given type */
    public abstract void emitPtrAssign(ILExpr ptr, ILType type, 
				       Location rhs_location)
	throws IOException;

    /** Generate code to read the contents of the given member of a
     * pointed-to record */
    public abstract Location emitFieldRead(ILExpr record_ptr, 
					   ILRecordMember member) 
	throws IOException;

    /** Generate code to assign to the given member of a pointed-to
     * record */
    public abstract void emitFieldAssign(ILExpr record_ptr,
					 ILRecordMember member,
					 Location rhs_location)
	throws IOException;

    /** Generate code to read the contents of the given sequence member
     * of a pointed-to record at the given index */
    public abstract Location emitSequenceFieldRead(ILExpr record_ptr, 
						   ILRecordMember member,
						   ILExpr index)
	throws IOException;

    /** Generate code to assign to the given sequence member of a
     * pointed-to record at the given index */
    public abstract void emitSequenceFieldAssign(ILExpr record_ptr, 
						 ILRecordMember member,
						 ILExpr index,
						 Location rhs_location)
	throws IOException;

    /** Generate code to move the address of a global variable into a
     * result register */
    public abstract Location emitGlobalAddress(String name) throws IOException;

    /** Generate code to move an integer constant into a result
     * register */
    public abstract Location emitIntConstant(int value) throws IOException;

    /** Generate an integer add instruction */
    public abstract Location emitIntAdd(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate an integer subtract instruction */
    public abstract Location emitIntSub(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate an integer multiply instruction */
    public abstract Location emitIntMul(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate an integer divide instruction */
    public abstract Location emitIntDiv(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate an integer negation instruction */
    public abstract Location emitIntNegate(ILExpr arg) throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than comparison into a result register */
    public abstract Location emitIntLessThanValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than comparison into a condition-code location */
    public abstract Location emitIntLessThanTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than-or-equal comparison into a result register */
    public abstract Location emitIntLessThanOrEqualValue(ILExpr arg1, 
							 ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than-or-equal comparison into a condition-code
     * location */
    public abstract Location emitIntLessThanOrEqualTest(ILExpr arg1,
							ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than-or-equal comparison into a result
     * register */
    public abstract Location emitIntGreaterThanOrEqualValue(ILExpr arg1,
							    ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than-or-equal comparison into a result
     * register */
    public abstract Location emitIntGreaterThanOrEqualTest(ILExpr arg1,
							   ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than comparison into a result register */
    public abstract Location emitIntGreaterThanValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than comparison into a condition-code location */
    public abstract Location emitIntGreaterThanTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer equal comparison into a result register */
    public abstract Location emitIntEqualValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer equal comparison into a condition-code location */
    public abstract Location emitIntEqualTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer not-equal comparison into a result register */
    public abstract Location emitIntNotEqualValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer not-equal comparison into a condition-code location */
    public abstract Location emitIntNotEqualTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * unsigned integer less-than comparison into a result register */
    public abstract Location emitUnsignedLessThanValue(ILExpr arg1, 
						       ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of an
     * unsigned integer less-than comparison into a condition-code
     * location */
    public abstract Location emitUnsignedLessThanTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the logical negation (0 if arg is
     * non-zero, non-zero (e.g. 1) if arg is 0) into a result
     * register */
    public abstract Location emitLogicalNegateValue(ILExpr arg) 
	throws IOException;

    /** Generate code to move the logical negation (0 if arg is
     * non-zero, non-zero (e.g. 1) if arg is 0) into a condition-code
     * location */
    public abstract Location emitLogicalNegateTest(ILExpr arg) 
	throws IOException;

    /** Generate code to test whether the integer argument represents
     * true (non-zero) */
    public abstract Location emitTest(ILExpr arg) throws IOException;

    /** Generate code to move a double-precision floating-point
     * constant into a result location */
    public abstract Location emitDoubleConstant(double value)
	throws IOException;

    /** Generate a double-precision floating-point add instruction */
    public abstract Location emitDoubleAdd(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate a double-precision floating-point subtract instruction */
    public abstract Location emitDoubleSub(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate a double-precision floating-point multiply instruction */
    public abstract Location emitDoubleMul(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate a double-precision floating-point divide instruction */
    public abstract Location emitDoubleDiv(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate a double-precision floating-point negation instruction */
    public abstract Location emitDoubleNegate(ILExpr arg) throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than comparison into a
     * result register */
    public abstract Location emitDoubleLessThanValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than comparison into a
     * condition-code location */
    public abstract Location emitDoubleLessThanTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than-or-equal comparison
     * into a result register */
    public abstract Location emitDoubleLessThanOrEqualValue(ILExpr arg1, 
							    ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than-or-equal comparison
     * into a condition-code location */
    public abstract Location emitDoubleLessThanOrEqualTest(ILExpr arg1,
							   ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than-or-equal
     * comparison into a result register */
    public abstract Location emitDoubleGreaterThanOrEqualValue(ILExpr arg1,
							       ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than-or-equal
     * comparison into a result register */
    public abstract Location emitDoubleGreaterThanOrEqualTest(ILExpr arg1,
							      ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than comparison into a
     * result register */
    public abstract Location emitDoubleGreaterThanValue(ILExpr arg1, 
							ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than comparison into a
     * condition-code location */
    public abstract Location emitDoubleGreaterThanTest(ILExpr arg1,
						       ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point equal comparison into a result
     * register */
    public abstract Location emitDoubleEqualValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point equal comparison into a
     * condition-code location */
    public abstract Location emitDoubleEqualTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point not-equal comparison into a
     * result register */
    public abstract Location emitDoubleNotEqualValue(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point not-equal comparison into a
     * condition-code location */
    public abstract Location emitDoubleNotEqualTest(ILExpr arg1, ILExpr arg2)
	throws IOException;

    /** Generate an instruction to convert an integer to a
     * double-precision floating-point number */
    public abstract Location emitIntToDouble(ILExpr arg) throws IOException;

    /** Generate code to allocate heap memory big enough to hold a
     * value of the given type */
    public abstract Location emitAllocate(ILType allocated_type) 
	throws IOException;

    /** Generate code to allocate heap memory big enough to hold a
     * value of the given type, whose last field is a sequence field
     * that should have length elements */
    public abstract Location emitAllocateSequence(ILType allocated_type, 
						  ILExpr length)
	throws IOException;

    /** Generate a direct call to a function */
    public abstract Location emitDirectCall(ILType result_type, 
					    String fun_name, 
					    List<ILExpr> args)
	throws IOException;

    /** Generate an indirect call through a function pointer */
    public abstract Location emitIndirectCall(ILType result_type, 
					      ILExpr fun_ptr, 
					      List<ILExpr> args)
	throws IOException;

    /** Generate a direct call to a runtime system function */
    public abstract Location emitRuntimeCall(ILType result_type, 
					     String fun_name, 
					     List<ILExpr> args)
	throws IOException;


    // Stack layout operations
    
    /** return the base offset in the enclosing function's stack frame
     * for local variables */
    public abstract int getLocalBaseOffset();

    /** return the base offset in the enclosing function's stack frame
     * for formal variables */
    public abstract int getFormalBaseOffset();


    // Type-specific size and alignment operations
    
    /** The size in bytes of an int value on this target */
    public abstract int intByteSize();

    /** The alignment constraint in bytes of an int value on this
     * target */
    public abstract int intByteAlign();
    
    /** The size in bytes of an int value on this target */
    public abstract int doubleByteSize();

    /** The alignment constraint in bytes of an int value on this
     * target */
    public abstract int doubleByteAlign();
    
    /** The size in bytes of an int value on this target */
    public abstract int ptrByteSize();

    /** The alignment constraint in bytes of an int value on this
     * target */
    public abstract int ptrByteAlign();


    /** A helper function to round up a byte offset to satisfy an
     * alignment constraint */
    public static int align(int offset, int align) {
	int excess = offset % align;
	if (excess != 0) {
	    // pad to next multiple of alignment
	    offset += align - excess;
	}
	return offset;
    }


    // Emitting helper operations

    /** helper to emit to the output file (printing if desired) */
    protected void emit(String output) throws IOException {
	output_file.write(output);
	if (printCode) {
	    // also print out the emitted code
	    System.out.print(output);
	}
    }

    /** helper to emit to the output file (printing if desired),
     * automatically appending a newline */
    protected void emitln(String output) throws IOException {
	emit(output);
	emit("\n");
    }
}
