/** This is the concrete class representing the X86 target machine. */

/*

This code generator does very simple instruction selection and
register allocation in a single pass.  All local variables are
allocated in fixed stack location, and loaded into registers only
during execution of a single statement.


** Data sizes:

Ints and pointers are 4 bytes, doubles are 8 bytes.


** Registers:

The X86 has 8 registers (that this code generator uses):

  EAX, EBX, ECX, and EDX are allocatable 4-byte registers

  ESI and EDI are held as scratch registers that can be overwritten by
  the code implementing any instruction

  EBP is the frame pointer (which always points to a fixed spot in
  each stack frame)

  ESP is the stack pointer (which always points to the end of the stack)

The X86 also has a floating-point stack on which doubles are stored
during most floating-point calculations.


** Procedure calling conventions:

The arguments to a procedure are pushed on the stack, right-to-left,
independent of type and alignment preferences.

The result of a procedure, if an integer or pointer, is in EAX, while
if a double, is on the top of the floating-point stack.


** Stack layout:

On the x86/linux, stack frames are layed out as follows (with stacks
growing down to lower addresses):

... caller's frame ...
-------------
formal arg n         [formals pushed by caller, accessible to callee]
formal arg n-1
...
formal arg 1
-------------
return address       [implicitly pushed by call instruction]
-------------
caller's frame ptr   [pushed/allocated by callee in its prologue]
		   <- frame pointer of callee points here, set in prologue
<space to save all caller-save registers at calls (4 of them)>
local variable m
local variable m-1
...
local variable 1
-------------
actual arg k         [actuals pushed by callee during a procedure call]
actual arg k-1
...
actual arg 1       <- stack pointer points at bottom of stack, always
-------------

*/


package CodeGen.X86;

import IL.*;
import CodeGen.*;
import Helpers.*;
import java.io.*;
import java.util.*;

public class X86Target extends Target {
    // Instance variables

    /** Whether or not commenting of each statement should be included
     * (for now, always do so) */
    protected boolean commentOutput = true;

    /** The name of the function being compiled, or null if not
     * in a function */
    protected String current_fun_name = null;

    /** The return type of the function being compiled, or null if not
     * in a function */
    protected ILType fun_return_type = null;

    /** (negative) offset relative to frame pointer of space to save
     * caller-save registers, if we're generating code for a
     * function */
    protected int callerSaveRegsBase;

    /** (negative) offset relative to frame pointer of first local
     * (and also the amount of space to add to the frame pointer to
     * allocate space for the frame), if we're generating code for a
     * function */
    protected int localsBase;

    /** (positive) offset relative to frame pointer of first formal,
     * if we're generating code for a function */
    protected int formalsBase;

    /** a helper used to generate unique labels */
    protected int label_number = 0;

    /** whether to do optimizations during target code emission */
    protected boolean optimize = false;


    /** Constructor */
    public X86Target() {
	// set up the register data structures
	initializeRegisters();
    }

    /** Enable optimizations (disabled by default) */
    public void enableOptimizations() {
	optimize = true;
    }


    // Code generation operations for program beginnings and endings

    /** Generate whatever code is needed by this target at the start
     * of the output file */
    public void emitProgramPrologue() throws IOException {
	// nothing needed
    }

    /** Generate whatever code is needed by this target at the end
     * of the output file */
    public void emitProgramEpilogue() throws IOException {
	// nothing needed
    }


    // Code generation operations for function beginnings and endings

    /** Generate code to declare and introduce a function, laying out
     * the function's stack frame in the process */
    public void emitFunctionPrologue(ILType return_type,
				     String fun_name,
				     List<ILFormalVarDecl> formals, 
				     Map<String,ILVarDecl> decls) 
		throws IOException {
	String compiled_fun_name;
	if (fun_name.equals("main")) {
	    // the main function is called from outside code to start
	    // the application, so it has to follow the outside
	    // system's naming conventions
	    compiled_fun_name = runtimeFunName(fun_name);
	} else {
	    // just use the regular function name
	    compiled_fun_name = fun_name;
	}
	
	// declare the function's global name
	emitOp(".text");
	emitln(".globl " + compiled_fun_name);
	emitOp(".align", "2");
	emitLabel(compiled_fun_name);

	// remember the function's name and return type
	current_fun_name = fun_name;
	fun_return_type = return_type;

	// compute the stack frame layout for this function
	layoutStackFrame(return_type, formals, decls);

	// save the caller's frame pointer onto the stack
	emitOp("pushl", regOperand(FP));

	// set the frame pointer from the stack pointer
	move(FP, SP);

	// subtract from the stack pointer to allocate space for the
	// caller-saved registers area and the local variables
	// (localsBase is negative, so we negate it to get the stack
	// size to allocate)
	emitOp("subl", intOperand(-localsBase), regOperand(SP));

	if (fun_name.equals("main")) {
	    // the regular C calling convention treats EBX as a
	    // callee-save register, so we need to preserve it
	    emitOp("pushl", regOperand(EBX));
	}
    }

    /** Generate code to finish a function */
    public void emitFunctionEpilogue(String fun_name)
		throws IOException {
	// nothing to generate
	// forget the function name and return type
	current_fun_name = null;
	fun_return_type = null;
    }


    // Code generation operations for global declarations

    /** Generate code to declare (but not initialize) a global
     * variable */
    public void emitUninitializedGlobalVarDecl(ILType type, String name)
		throws IOException {
	// generate the declaration itself
	emitGlobalVarDecl(type, name);
	// leave space big enough to hold the value
	emitOp(".skip", Integer.toString(type.byteSize(this)));
    }

    /** Generate code to declare and initialize a global variable */
    public void emitInitializedGlobalVarDecl(ILType type, String name,
					     ILValue value)
		throws IOException {
	// generate the declaration itself
	emitGlobalVarDecl(type, name);
	// generate the initial data
	value.codegenInitializer(this);
    }

    /** Generate code to declare a global variable */
    protected void emitGlobalVarDecl(ILType type, String name)
		throws IOException {
	emitOp(".data");
	emitln(".globl " + name);
	emitAlignDirective(type);
	emitLabel(name);
    }

    /** Generate code to initialize a global memory location to a
     * double */
    public void emitAlignDirective(ILType type) throws IOException {
	emitOp(".align", Integer.toString(type.byteAlign(this)));
    }

    /** Generate code to initialize a global memory location to an
     * int */
    public void emitIntInitializer(int value) throws IOException {
	emitOp(".long", Integer.toString(value));
    }

    /** Generate code to initialize a global memory location to a
     * double */
    public void emitDoubleInitializer(double value) throws IOException {
	emitOp(".double", Double.toString(value));
    }

    /** Generate code to initialize a global memory location to the
     * address of the global variable with the given name */
    public void emitGlobalAddressInitializer(String name) throws IOException {
	emitOp(".long", name);
    }

    /** Generate code to initialize a global memory location to the
     * address of a label */
    public void emitLabelAddressInitializer(ILLabel label) throws IOException {
	emitOp(".long", label.getName());
    }


    // Code generation operations for each of the different kinds of
    // statements in the IL

    /** Generate code for an assignment */
    public void emitAssign(ILAssignableExpr lhs, ILExpr rhs)
		throws IOException {
	if (commentOutput) {
	    emitComment(lhs.toString() + " := " + rhs.toString() + ";");
	}

	// generate code for the rhs into a register
	Location rhs_location = rhs.codegen(this);

	// then generate the assignment to the lhs location
	lhs.codegenAssign(rhs_location, this);

	// deallocate the rhs expression's register
	deallocateRegister(rhs_location);
    }

    /** Generate code for a zero/false conditional branch */
    public void emitConditionalBranchFalse(ILExpr test, ILLabel target_label)
		throws IOException {
	if (commentOutput) {
	    emitComment("iftrue " + test.toString() + " goto " + 
			target_label.toString() + ";");
	}

	// generate code for the test, as a test that sets condition codes
	Location test_location = test.codegenTest(this);

	// verify that this is a condition code location
	if (! (test_location instanceof X86ComparisonResultLocation)) {
	    throw new InternalCompilerError("expecting a condition code");
	}
	X86ComparisonResultLocation test_cc =
	    (X86ComparisonResultLocation)test_location;

	// emit the branch is-zero to the label
	emitOp("j" + test_cc.branchFalseOp(), target_label.getName());
    }

    /** Generate code for a non-zero/true conditional branch */
    public void emitConditionalBranchTrue(ILExpr test, ILLabel target_label)
		throws IOException {
	if (commentOutput) {
	    emitComment("iftrue " + test.toString() + " goto " + 
			target_label.toString() + ";");
	}

	// generate code for the test, as a test that sets condition codes
	Location test_location = test.codegenTest(this);

	// verify that this is a condition code location
	if (! (test_location instanceof X86ComparisonResultLocation)) {
	    throw new InternalCompilerError("expecting a condition code");
	}
	X86ComparisonResultLocation test_cc =
	    (X86ComparisonResultLocation)test_location;

	// emit the branch non-zero to the label
	emitOp("j" + test_cc.branchTrueOp(), target_label.getName());
    }

    /** Generate code for an expression, ignoring the result */
    public void emitExpr(ILExpr expr) throws IOException {
	if (commentOutput) {
	    emitComment(expr.toString() + ";");
	}

	// evaluate the expression
	Location result_location = expr.codegen(this);

	// ignore result
	deallocateRegister(result_location);
    }

    /** Generate code for a goto statement */
    public void emitGoto(ILLabel label) throws IOException {
	if (commentOutput) {
	    emitComment("goto " + label.toString() + ";");
	}

	emitOp("jmp", label.getName());

	// does not fall through
    }

    /** Generate code for a label */
    public void emitLabel(ILLabel label) throws IOException {
	if (commentOutput) {
	    emitComment("label " + label.toString() + ";");
	}

	emitLabel(label.getName());
    }

    /** Generate code for a return */
    public void emitReturn(ILExpr result) throws IOException {
	if (commentOutput) {
	    emitComment("return " + result.toString() + ";");
	}

	// evaluate the expression
	Location result_location = result.codegen(this);

	// move the result into the return register defined by the
	// calling convention
	Location return_location = returnLocation(fun_return_type);
	move(return_location, result_location);

	// deallocate the result register
	deallocateRegister(result_location);

	if (current_fun_name.equals("main")) {
	    // the regular C calling convention treats EBX as a
	    // callee-save register, so we need to restore it
	    emitOp("popl", regOperand(EBX));
	}

	// the leave instruction restores the caller's saved frame
	// pointer, which implicitly cuts back the stack to just the
	// return address
	emitOp("leave");
	
	// do the return, now that the return address is on top of the
	// stack
	emitOp("ret");

	// (the caller is responsible for popping off its arguments)

	// doesn't fall through
    }

    /** Generate code to signal an error */
    public void emitThrowException(String message) throws IOException {
	if (commentOutput) {
	    emitComment("throw new Exception(\"" + message + "\");");
	}

	// allocate a label for the string data
	String string_label = getNewLabel();

	// declare the string
	emitOp(".section", ".rodata");
	emitLabel(string_label);
	emitOp(".string", "\"" + message + "\"");
	emitOp(".text");
	
	// push its address (but can't use push instruction directly
	// on an address operand)
	emitOp("subl", intOperand(4), regOperand(SP));
	emitOp("movl", 
	       globalAddressOperand(string_label), 
	       ptrDerefOperand(SP));

	// call a runtime system helper
	emitOp("call", runtimeFunName("throw_exception"));

	// does not fall through
    }


    // Code generation operations for each of the different kinds of
    // expressions in the IL

    /** Generate code to read the contents of a local variable */
    public Location emitVarRead(ILVarDecl decl) throws IOException {
	// local variables are allocated on the stack, and accessed at
	// a fixed offset off of the frame pointer.  stack frame
	// layout was done when the function prologue was generated.

	// get the type of the variable
	ILType type = decl.getType();

	// compute the offset of the variable in the stack frame
	int offset = decl.getByteOffset(this);

	// allocate a register for the result
	Location result_location = allocateRegister(type);

	// generate a load off of the frame pointer
	emitLoadOffset(result_location, FP, offset, type);

	// return where the result is
	return result_location;
    }

    /** Generate code to assign to a local variable */
    public void emitVarAssign(ILVarDecl decl, Location rhs_location)
		throws IOException {
	// local variables are allocated on the stack, and accessed at
	// a fixed offset off of the frame pointer.  stack frame
	// layout was done when the function prologue was generated.

	// get the type of the variable
	ILType type = decl.getType();

	// compute the offset of the variable in the stack frame
	int offset = decl.getByteOffset(this);

	// generate a store off of the frame pointer
	emitStoreOffset(FP, offset, type, rhs_location);
    }

    /** Generate code to read the target of a pointer that points to a
     * value of the given type */
    public Location emitPtrRead(ILExpr ptr, ILType type) throws IOException {
	// evaluate the pointer
	Location ptr_location = ptr.codegen(this);

	// deallocate the pointer's register, then allocate a register
	// for the result (can deallocate since the result won't be
	// written to until the pointer's register is finished being
	// used)
	deallocateRegister(ptr_location);
	Location result_location = allocateRegister(type);

	// generate a load
	emitLoad(result_location, ptr_location, type);

	// return where the result is
	return result_location;
    }

    /** Generate code to assign to the target of a pointer that points
     * to a value of the given type */
    public void emitPtrAssign(ILExpr ptr, ILType type, 
			      Location rhs_location)
		throws IOException {
	// evaluate the pointer
	Location ptr_location = ptr.codegen(this);

	// generate a store
	emitStore(ptr_location, type, rhs_location);

	// deallocate the pointer register
	deallocateRegister(ptr_location);
    }

    /** Generate code to read the contents of the given member of a
     * pointed-to record */
    public Location emitFieldRead(ILExpr record_ptr, ILRecordMember member) 
		throws IOException {
	// evaluate the record pointer
	Location record_ptr_location = record_ptr.codegen(this);

	// compute the type of the member
	ILType type = member.getType();

	// compute the member's offset
	int offset = member.getByteOffset();

	// deallocate the pointer's register, then allocate a register
	// for the result (can deallocate since the result won't be
	// written to until the pointer's register is finished being
	// used)
	deallocateRegister(record_ptr_location);
	Location result_location = allocateRegister(type);

	// generate a load
	emitLoadOffset(result_location, record_ptr_location, offset, type);

	// return where the result is
	return result_location;
    }

    /** Generate code to assign to the given member of a pointed-to
     * record */
    public void emitFieldAssign(ILExpr record_ptr, ILRecordMember member,
				Location rhs_location)
		throws IOException {
	// evaluate the record pointer
	Location record_ptr_location = record_ptr.codegen(this);

	// compute the type of the member
	ILType type = member.getType();

	// compute the member's offset
	int offset = member.getByteOffset();

	// generate a store
	emitStoreOffset(record_ptr_location, offset, type, rhs_location);

	// deallocate the record pointer register
	deallocateRegister(record_ptr_location);
    }

    /** Generate code to read the contents of the given sequence member
     * of a pointed-to record at the given index */
    public Location emitSequenceFieldRead(ILExpr record_ptr, 
					  ILRecordMember member,
					  ILExpr index)
		throws IOException {
	// evaluate the record pointer
	Location record_ptr_location = record_ptr.codegen(this);

	// evaluate the index
	Location index_location = index.codegen(this);

	// compute the type of the member
	ILType type = member.getType().getSequenceElemType();

	// compute the member's offset
	int offset = member.getByteOffset();

	// deallocate the record pointer's and the index's registers,
	// then allocate a register for the result (can deallocate
	// since the result won't be written to until the pointer's
	// and the index's registers are finished being used)
	deallocateRegister(record_ptr_location);
	deallocateRegister(index_location);
	Location result_location = allocateRegister(type);

	// generate a load
	emitLoadOffsetSequence(result_location, record_ptr_location, offset, 
			       index_location, type);

	// return where the result is
	return result_location;
    }

    /** Generate code to assign to the given sequence member of a
     * pointed-to record at the given index */
    public void emitSequenceFieldAssign(ILExpr record_ptr, 
					ILRecordMember member,
					ILExpr index,
					Location rhs_location)
		throws IOException {
	// evaluate the record pointer
	Location record_ptr_location = record_ptr.codegen(this);

	// evaluate the index
	Location index_location = index.codegen(this);

	// compute the type of the member
	ILType type = member.getType().getSequenceElemType();

	// compute the member's offset
	int offset = member.getByteOffset();

	// generate a store
	emitStoreOffsetSequence(record_ptr_location, offset, index_location, 
				type, rhs_location);

	// deallocate the record pointer and index registers
	deallocateRegister(record_ptr_location);
	deallocateRegister(index_location);
    }

    /** Generate code to move the address of a global variable into a
     * result register */
    public Location emitGlobalAddress(String name) throws IOException {
	// compute the pointer type (actually, we don't know, nor care
	// for now, what the type of the global is, only that the
	// result is a pointer.  so use int as a placeholder type for
	// the pointer target)
	ILType type = new ILPtrType(new ILIntType());

	// allocate a register for the result
	Location result_location = allocateRegister(type);

	// generate the instruction
	emitOp("movl", 
	       globalAddressOperand(name), 
	       regOperand(result_location));

	// return where the result is
	return result_location;
    }

    /** Generate code to move an integer constant into a result
     * register */
    public Location emitIntConstant(int value) throws IOException {
	// allocate a register for the result
	Location result_location = allocateRegister(new ILIntType());

	// generate the instruction
	emitOp("movl", intOperand(value), regOperand(result_location));

	// return where the result is
	return result_location;
    }

    /** Generate an integer add instruction */
    public Location emitIntAdd(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleIntBinop(new ILIntType(), "add", arg1, arg2);
    }

    /** Generate an integer subtract instruction */
    public Location emitIntSub(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleIntBinop(new ILIntType(), "sub", arg1, arg2);
    }

    /** Generate an integer multiply instruction */
    public Location emitIntMul(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleIntBinop(new ILIntType(), "imul", arg1, arg2);
    }

    /** Generate an integer divide instruction */
    public Location emitIntDiv(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate the arguments
	Location arg1_location = arg1.codegen(this);
	Location arg2_location = arg2.codegen(this);
	
	// deallocate the arguments' registers, then allocate a
	// register for the result, preferring where the result will
	// be computed naturally (can deallocate since the result
	// won't be written to until the arguments' registers are
	// finished being used)
	deallocateRegister(arg1_location);
	deallocateRegister(arg2_location);
	Location result_location = allocateRegister(new ILIntType(), EAX);

	// the divide instruction requires its dividend (arg1) to be
	// sign-extended in EAX and EDX, and the resulting quotient
	// and remainder are put into EAX and EDX, respectively.  So
	// we have to free up these two registers, if the source isn't
	// already in them, and move the result out, if it's not
	// already wanted in them.

	// save EDX into a scratch register, if is being used for some
	// other needed value.
	boolean EDX_saved = 
	    (EDX == arg2_location || 
	     isAllocated(EDX) && (EDX != result_location));
	if (EDX_saved) {
	    move(SCRATCH1, EDX);
	    if (arg2_location == EDX) {
		// update the location of arg2
		arg2_location = SCRATCH1;
	    }
	}

	// save EAX into a scratch register, if is being used for some
	// other needed value.
	boolean EAX_saved = 
	    (EAX == arg2_location || 
	     isAllocated(EAX) && (EAX != result_location));
	if (EAX_saved) {
	    move(SCRATCH2, EAX);
	    if (arg2_location == EAX) {
		// update the location of arg2
		arg2_location = SCRATCH2;
	    }
	}

	// save EAX into a scratch register, if it's not the result
	// register (and so not needed)
	if (arg2_location == EAX) {
	    move(SCRATCH2, EAX);
	    arg2_location = SCRATCH2;
	}

	// move the dividend into EAX, if needed
	move(EAX, arg1_location);

	// sign-extend the dividend (which reads EAX and writes to EAX
	// and EDX)
	emitOp("cdq");

	// do the division (which reads EAX/EDX and its operand (the
	// divisor) and writes EAX (the quotient) & EDX (the remainder))
	emitOp("idivl", regOperand(arg2_location));

	// move the computed result into the chosen result register
	move(result_location, EAX);

	// restore each of EDX and EAX from the scratch registers,
	// unless it's the result register (so that we overwrote it
	// anyway)
	if (EDX_saved && (EDX != result_location)) move(EDX, SCRATCH1);
	if (EAX_saved && (EAX != result_location)) move(EAX, SCRATCH2);

	// return the register holding the quotient
	return result_location;
    }

    /** Generate an integer negation instruction */
    public Location emitIntNegate(ILExpr arg) throws IOException {
	return emitSimpleIntUnop(new ILIntType(), "neg", arg);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than comparison into a result register */
    public Location emitIntLessThanValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("l", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than comparison into a condition-code location */
    public Location emitIntLessThanTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("l", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than-or-equal comparison into a result register */
    public Location emitIntLessThanOrEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("le", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer less-than-or-equal comparison into a condition-code
     * location */
    public Location emitIntLessThanOrEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("le", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than-or-equal comparison into a result
     * register */
    public Location emitIntGreaterThanOrEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("ge", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than-or-equal comparison into a result
     * register */
    public Location emitIntGreaterThanOrEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("ge", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than comparison into a result register */
    public Location emitIntGreaterThanValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("g", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer greater-than comparison into a condition-code location */
    public Location emitIntGreaterThanTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("g", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer equal comparison into a result register */
    public Location emitIntEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("e", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer equal comparison into a condition-code location */
    public Location emitIntEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("e", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer not-equal comparison into a result register */
    public Location emitIntNotEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareValue("ne", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * integer not-equal comparison into a condition-code location */
    public Location emitIntNotEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitIntCompareTest("ne", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * unsigned integer less-than comparison into a result register */
    public Location emitUnsignedLessThanValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// b is unsigned l
	return emitIntCompareValue("b", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of an
     * unsigned integer less-than comparison into a condition-code
     * location */
    public Location emitUnsignedLessThanTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// b is unsigned l
	return emitIntCompareTest("b", arg1, arg2);
    }

    /** Generate code to move the logical negation (0 if arg is
     * non-zero, non-zero (e.g. 1) if arg is 0) into a result
     * register */
    public Location emitLogicalNegateValue(ILExpr arg) throws IOException {
	// this is the same as arg == 0

	// evaluate the argument as a boolean condition
	Location arg_location = arg.codegenTest(this);

	// verify that this is a condition code location
	if (! (arg_location instanceof X86ComparisonResultLocation)) {
	    throw new InternalCompilerError("expecting a condition code");
	}
	X86ComparisonResultLocation arg_cc =
	    (X86ComparisonResultLocation)arg_location;

	// emit code to set the boolean result
	return emitBooleanValue(arg_cc.branchFalseOp());
    }

    /** Generate code to move the logical negation (0 if arg is
     * non-zero, non-zero (e.g. 1) if arg is 0) into a condition-code
     * location */
    public Location emitLogicalNegateTest(ILExpr arg) throws IOException {
	// this is the same as arg == 0

	// evaluate the argument as a boolean condition
	Location arg_location = arg.codegenTest(this);

	// verify that this is a condition code location
	if (! (arg_location instanceof X86ComparisonResultLocation)) {
	    throw new InternalCompilerError("expecting a condition code");
	}
	X86ComparisonResultLocation arg_cc =
	    (X86ComparisonResultLocation)arg_location;

	// return the negated condition code result
	return arg_cc.negate();
    }

    /** Generate code to test whether the argument represents true */
    public Location emitTest(ILExpr arg) throws IOException {
	// evaluate the argument
	Location arg_location = arg.codegen(this);

	// compare against 0 (false)
	// (the compare instruction compares its arguments in the
	// reverse order!)
	emitOp("cmpl", intOperand(0), regOperand(arg_location));
	
	// deallocate the argument register
	deallocateRegister(arg_location);
	
	// return the condition code for a not-equal test
	return new X86ComparisonResultLocation("ne");
    }


    /** Generate code to move a double-precision floating-point
     * constant into a result location */
    public Location emitDoubleConstant(double value) throws IOException {
	// a double constant needs to be declared as initialized
	// global data, then loaded onto the floating point stack

	// allocate a label for the double data
	String double_label = getNewLabel();

	// declare the double
	emitOp(".section", ".rodata");
	emitOp(".align", "8");
	emitLabel(double_label);
	emitOp(".double", Double.toString(value));
	emitOp(".text");
	
	// move its address into a scratch register
	emitOp("movl", 
	       globalAddressOperand(double_label), 
	       regOperand(SCRATCH1));

	// load its contents onto the floating point stack
	emitLoad(FPS, SCRATCH1, new ILDoubleType());

	// return where the result is
	return FPS;
    }

    /** Generate a double-precision floating-point add instruction */
    public Location emitDoubleAdd(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleDoubleBinop("add", arg1, arg2);
    }

    /** Generate a double-precision floating-point subtract instruction */
    public Location emitDoubleSub(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleDoubleBinop("sub", arg1, arg2);
    }

    /** Generate a double-precision floating-point multiply instruction */
    public Location emitDoubleMul(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleDoubleBinop("mul", arg1, arg2);
    }

    /** Generate a double-precision floating-point divide instruction */
    public Location emitDoubleDiv(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitSimpleDoubleBinop("div", arg1, arg2);
    }

    /** Generate a double-precision floating-point negation instruction */
    public Location emitDoubleNegate(ILExpr arg) throws IOException {
	return emitSimpleDoubleUnop("chs", arg);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than comparison into a
     * result register */
    public Location emitDoubleLessThanValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// b is unsigned/double l
	return emitDoubleCompareValue("b", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than comparison into a
     * condition-code location */
    public Location emitDoubleLessThanTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// b is unsigned/double l
	return emitDoubleCompareTest("b", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than-or-equal comparison
     * into a result register */
    public Location emitDoubleLessThanOrEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// be is unsigned/double le
	return emitDoubleCompareValue("be", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point less-than-or-equal comparison
     * into a condition-code location */
    public Location emitDoubleLessThanOrEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// be is unsigned/double le
	return emitDoubleCompareTest("be", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than-or-equal
     * comparison into a result register */
    public Location emitDoubleGreaterThanOrEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// ae is unsigned/double ge
	return emitDoubleCompareValue("ae", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than-or-equal
     * comparison into a result register */
    public Location emitDoubleGreaterThanOrEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// ae is unsigned/double ge
	return emitDoubleCompareTest("ae", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than comparison into a
     * result register */
    public Location emitDoubleGreaterThanValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// a is unsigned/double g
	return emitDoubleCompareValue("a", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point greater-than comparison into a
     * condition-code location */
    public Location emitDoubleGreaterThanTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// a is unsigned/double g
	return emitDoubleCompareTest("a", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point equal comparison into a result
     * register */
    public Location emitDoubleEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitDoubleCompareValue("e", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point equal comparison into a
     * condition-code location */
    public Location emitDoubleEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitDoubleCompareTest("e", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point not-equal comparison into a
     * result register */
    public Location emitDoubleNotEqualValue(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitDoubleCompareValue("ne", arg1, arg2);
    }

    /** Generate code to move the result (0=false, non-0=true) of a
     * double-precision floating-point not-equal comparison into a
     * condition-code location */
    public Location emitDoubleNotEqualTest(ILExpr arg1, ILExpr arg2)
		throws IOException {
	return emitDoubleCompareTest("ne", arg1, arg2);
    }

    /** Generate an instruction to convert an integer to a
     * double-precision floating-point number */
    public Location emitIntToDouble(ILExpr arg) throws IOException {
	// evaluate the argument
	Location arg_location = arg.codegen(this);
	
	// push the integer onto the floating-point stack, converting
	// it to floating-point format in the process.  but we can't
	// move directly from a register to the floating-point stack,
	// so we first push it onto the stack, then move it, then pop
	// the stack.
	emitOp("pushl", regOperand(arg_location));
	emitOp("fildl", ptrDerefOperand(SP));
	emitOp("addl", intOperand(4), regOperand(SP));

	// deallocate the argument register
	deallocateRegister(arg_location);
	
	// return where the result is
	return FPS;
    }


    /** Generate code to allocate heap memory big enough to hold a
     * value of the given type */
    public Location emitAllocate(ILType allocated_type) throws IOException {
	// push the number of bytes to allocate
	emitOp("pushl", intOperand(allocated_type.byteSize(this)));

	// save any live registers
	saveRegisters();

	// call a runtime system helper
	emitOp("call", runtimeFunName("allocate"));

	// pop off the one 4-byte argument
	emitOp("addl", intOperand(4), regOperand(SP));

	// allocate a register for the result, preferring where the
	// call comes back naturally
	ILType result_type = new ILPtrType(allocated_type);
	Location return_location = returnLocation(result_type);
	Location result_location = 
	    allocateRegister(result_type, return_location);

	// move where the return value comes back to its chosen register
	move(result_location, return_location);

	// restore any live registers, except the result register
	restoreRegisters(result_location);

	// return where the result is
	return result_location;
    }

    /** Generate code to allocate heap memory big enough to hold a
     * value of the given type, whose last field is a sequenced field
     * that should have length elements */
    public Location emitAllocateSequence(ILType allocated_type, ILExpr length)
		throws IOException {
	// evaluate the length
	Location length_location = length.codegen(this);

	// push the arguments on, right-to-left

	// push the size of each sequence element
	emitOp("pushl", 
	       intOperand(allocated_type.sequenceElementByteSize(this)));

	// push the length
	emitOp("pushl", regOperand(length_location));

	// push the number of non-sequence bytes to allocate
	emitOp("pushl", intOperand(allocated_type.byteSize(this)));

	// deallocate the length register
	deallocateRegister(length_location);

	// save any live registers
	saveRegisters();

	// call a runtime system helper
	emitOp("call", runtimeFunName("allocate_sequence"));

	// pop off the 3 4-byte arguments
	emitOp("addl", intOperand(3*4), regOperand(SP));

	// allocate a register for the result, preferring where the
	// call comes back naturally
	ILType result_type = new ILPtrType(allocated_type);
	Location return_location = returnLocation(result_type);
	Location result_location = 
	    allocateRegister(result_type, return_location);

	// move where the return value comes back to its chosen register
	move(result_location, return_location);

	// restore any live registers, except the result register
	restoreRegisters(result_location);

	// return where the result is
	return result_location;
    }

    /** Generate a direct call to a function */
    public Location emitDirectCall(ILType result_type, String fun_name, 
				   List<ILExpr> args)
		throws IOException {
	// evaluate and push the arguments onto the stack
	int arg_stack_space = emitCallArguments(args);

	// save any live registers
	saveRegisters();

	// call the callee
	emitOp("call", fun_name);

	// pop off the arguments
	emitOp("addl", intOperand(arg_stack_space), regOperand(SP));

	// allocate a register for the result, preferring where the
	// call comes back naturally
	Location return_location = returnLocation(result_type);
	Location result_location = 
	    allocateRegister(result_type, return_location);

	// move where the return value comes back to its chosen register
	move(result_location, return_location);

	// restore any live registers, except the result register
	restoreRegisters(result_location);

	// return where the result is
	return result_location;
    }

    /** Generate an indirect call through a function pointer */
    public Location emitIndirectCall(ILType result_type, ILExpr fun_ptr, 
				     List<ILExpr> args)
		throws IOException {
	// evaluate and push the arguments onto the stack
	int arg_stack_space = emitCallArguments(args);

	// compute the function pointer
	Location fun_ptr_location = fun_ptr.codegen(this);

	// deallocate the function pointer's register
	deallocateRegister(fun_ptr_location);

	// save any live registers
	saveRegisters();

	// call the callee
	emitOp("call", "*" + regOperand(fun_ptr_location));

	// pop off the arguments
	emitOp("addl", intOperand(arg_stack_space), regOperand(SP));

	// allocate a register for the result, preferring where the
	// call comes back naturally
	Location return_location = returnLocation(result_type);
	Location result_location = 
	    allocateRegister(result_type, return_location);

	// move where the return value comes back to its chosen register
	move(result_location, return_location);

	// restore any live registers, except the result register
	restoreRegisters(result_location);

	// return where the result is
	return result_location;
    }

    /** Generate a direct call to a runtime system function */
    public Location emitRuntimeCall(ILType result_type, String fun_name, 
				    List<ILExpr> args)
		throws IOException {
	// same as a direct call to a program function
	return emitDirectCall(result_type, runtimeFunName(fun_name), args);
    }


    // Code generation helpers

    /** A helper that returns the way to refer to a register operand */
    protected String regOperand(Location location) {
	// this had better be a register
	if (! (location instanceof X86Register)) {
	    throw new InternalCompilerError("expecting a register");
	}
	X86Register reg = (X86Register)location;
	// return its way of being referenced in 
	return "%" + reg.getName();
    }

    /** A helper that returns the way to refer to the address of a
     * global variable as an operand */
    protected String globalAddressOperand(String name) {
	return "$" + name;
    }

    /** A helper that returns the way to refer to an integer constant
     * operand */
    protected String intOperand(int value) {
	return "$" + Integer.toString(value);
    }

    /** A helper that returns the way to refer to a pointer
     * dereference operand */
    protected String ptrDerefOperand(Location ptr_location) {
	return "(" + regOperand(ptr_location) + ")";
    }

    /** A helper that returns the way to refer to a pointer + offset
     * dereference operand */
    protected String ptrOffsetDerefOperand(Location ptr_location, int offset) {
	return Integer.toString(offset) + "(" + regOperand(ptr_location) + ")";
    }

    /** A helper that returns the way to refer to a pointer + offset +
     * index*sizeof(type) dereference operand */
    protected String ptrOffsetSequenceDerefOperand(Location ptr_location,
						   int offset,
						   Location index_location, 
						   ILType type) {
	return Integer.toString(offset) + 
	    "(" + regOperand(ptr_location) + "," +
	    regOperand(index_location) + "," + 
	    Integer.toString(type.byteSize(this)) + ")";
    }

    /** A helper to emit an operation with no arguments. */
    protected void emitOp(String op) throws IOException {
	emitln("\t" + op);
    }

    /** A helper to emit an operation with one argument. */
    protected void emitOp(String op, String arg) throws IOException {
	emitOp(op + "\t" + arg);
    }

    /** A helper to emit an operation with two arguments. */
    protected void emitOp(String op, String arg1, String arg2)
		throws IOException {
	emitOp(op, arg1 + ", " + arg2);
    }

    /** A helper to emit a label. */
    protected void emitLabel(String label) throws IOException {
	emitln(label + ":");
    }

    /** A helper to emit a comment. */
    protected void emitComment(String comment) throws IOException {
	// note that this doesn't work if the comment itself contains */!
	emitln("\t/* " + comment + " */");
    }

    /** A helper to move src into result, if they're different. */
    protected void move(Location result_location, Location src_location) 
		throws IOException {
	if (result_location != src_location) {
	    if (result_location == FPS ||
		src_location == FPS) {
		// moving from register to floating-point stack or
		// vice versa; weird
		throw new InternalCompilerError(
		    "unexpected move between data & floating-point registers");
	    }
	    emitOp("movl", 
		   regOperand(src_location),
		   regOperand(result_location));
	}
    }

    /** A helper to generate a normal integer binary 2-address
     * operation */
    protected Location emitSimpleIntBinop(ILType result_type, String opString,
					  ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate the arguments
	Location arg1_location = arg1.codegen(this);
	Location arg2_location = arg2.codegen(this);
	
	// emit the 2-address code
	emitOp(opString + "l", 
	       regOperand(arg2_location), regOperand(arg1_location));

	// this machine's 2-operand format computes the result into
	// the same register as arg1, and it's fine to just choose
	// that as the result register, since we're about to
	// deallocate it anyway

	// deallocate the other register
	deallocateRegister(arg2_location);

	// return where the result is
	return arg1_location;
    }

    /** A helper to generate a normal integer unary 1-address
     * operation */
    protected Location emitSimpleIntUnop(ILType result_type, String opString,
					 ILExpr arg)
		throws IOException {
	// evaluate the argument
	Location arg_location = arg.codegen(this);
	
	// emit the 1-address code
	emitOp(opString + "l", regOperand(arg_location));

	// this machine's 1-operand format computes the result into
	// the same register as the arg, and it's fine to just choose
	// that as the result register, since we're about to
	// deallocate it anyway

	// return where the result is
	return arg_location;
    }

    /** A helper to generate code to compute the result (0=false,
     * non-0=true) of an integer comparison as a value */
    protected Location emitIntCompareValue(String cmpOp, 
					   ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate and compare the arguments
	emitIntCompare(arg1, arg2);

	// emit code to compute the boolean result value
	return emitBooleanValue(cmpOp);
    }

    /** A helper that generates code to compare the integer arguments,
     * returning a condition-code location recording which codes to
     * use */
    protected Location emitIntCompareTest(String cmpOp, 
					  ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate and compare the arguments
	emitIntCompare(arg1, arg2);

	// return the location representing the condition codes to test
	return new X86ComparisonResultLocation(cmpOp);
    }

    /** A helper that generates code to compare the integer arguments,
     * setting the condition codes */
    protected void emitIntCompare(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate the arguments
	Location arg1_location = arg1.codegen(this);
	Location arg2_location = arg2.codegen(this);

	// the compare instruction compares its arguments in the
	// reverse order!
	emitOp("cmpl", 
	       regOperand(arg2_location), regOperand(arg1_location));
	
	// deallocate the argument registers
	deallocateRegister(arg1_location);
	deallocateRegister(arg2_location);
    }

    /** A helper to generate a normal double-precision floating-point
     * binary 2-address operation */
    protected Location emitSimpleDoubleBinop(String opString,
					     ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate the arguments onto the floating-point stack
	// (despite the Intel documentation, it appears that the first
	// operand should be on the top of the stack, with the second
	// argument below it, so evaluate the arguments in the reverse
	// order)
	Location arg2_location = arg2.codegen(this);
	Location arg1_location = arg1.codegen(this);
	
	// emit the floating point operation, which pops off the
	// arguments and pushes the result
	emitOp("f" + opString + "p");

	// return where the result is
	return FPS;
    }

    /** A helper to generate a normal double-precision floating-point
     * unary 1-address operation */
    protected Location emitSimpleDoubleUnop(String opString, ILExpr arg)
		throws IOException {
	// evaluate the argument onto the floating-point stack
	Location arg_location = arg.codegen(this);
	
	// emit the floating point operation, which operates directly
	// on the value on the top of the stack
	emitOp("f" + opString);

	// return where the result is
	return FPS;
    }

    /** A helper to generate code to compute the result (0=false,
     * non-0=true) of a double-precision floating-point comparison as
     * a value */
    protected Location emitDoubleCompareValue(String cmpOp, 
					      ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate and compare the arguments
	emitDoubleCompare(arg1, arg2);

	// emit code to compute the boolean result value
	return emitBooleanValue(cmpOp);
    }

    /** A helper that generates code to compare the double-precision
     * floating-point arguments, returning a condition-code location
     * recording which codes to use */
    protected Location emitDoubleCompareTest(String cmpOp, 
					     ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate and compare the arguments
	emitDoubleCompare(arg1, arg2);

	// return the location representing the condition codes to test
	return new X86ComparisonResultLocation(cmpOp);
    }

    /** A helper that generates code to compare the double-precision
     * floating-point arguments, setting the condition codes */
    protected void emitDoubleCompare(ILExpr arg1, ILExpr arg2)
		throws IOException {
	// evaluate the arguments onto the floating-point stack
	// (the compare instruction compares its arguments "backwards",
	// so evaluate the arguments in the reverse order)
	Location arg2_location = arg2.codegen(this);
	Location arg1_location = arg1.codegen(this);

	// we're going to overwrite AX, so save EAX to a scratch
	// register if allocated
	if (EAX.isAllocated()) {
	    move(SCRATCH1, EAX);
	}

	// compare the values on the top of the stack, popping off the
	// arguments and setting the floating-point condition codes
	emitOp("fcompp");

	// move the floating-point status word to the low half of the
	// EAX data register
	emitOp("fstsw", "%ax");
	
	// copy the upper-half byte of AX into the data condition
	// codes
	emitOp("sahf");

	// restore EAX, if we saved it
	if (EAX.isAllocated()) {
	    move(EAX, SCRATCH1);
	}
    }

    /** A helper to generate a value (0=false, non-0=true) based on
     * the condition codes and the given comparison operator */
    protected Location emitBooleanValue(String cmpOp) throws IOException {
	// allocate a register for the result
	Location result_location = allocateRegister(new ILIntType());

	// test the comparison outcome
	String true_label = getNewLabel();
	emitOp("j" + cmpOp, true_label);

	// fall through => test is false

	// return 0 (false)
	emitOp("movl", intOperand(0), regOperand(result_location));

	String done_label = getNewLabel();
	emitOp("jmp", done_label);

	emitLabel(true_label);

	// taken branch => test is true

	// return 1 (true)
	emitOp("movl", intOperand(1), regOperand(result_location));

	emitLabel(done_label);

	// return where the result is
	return result_location;
    }

    /** A helper that generates a pointer load instruction of data of
     * the size of the given type */
    protected void emitLoad(Location result_location, Location ptr_location,
			    ILType type)
		throws IOException {
	if (result_location == FPS) {
	    // load and push onto floating-point stack
	    emitOp("fldl", ptrDerefOperand(ptr_location));
	} else {
	    // load into target register
	    emitOp("movl", 
		   ptrDerefOperand(ptr_location),
		   regOperand(result_location));
	}
    }

    /** A helper that generates a pointer store instruction of data of
     * the size of the given type */
    protected void emitStore(Location ptr_location, ILType type,
			     Location arg_location)
		throws IOException {
	if (arg_location == FPS) {
	    // pop off floating-point stack and store
	    emitOp("fstpl", ptrDerefOperand(ptr_location));
	} else {
	    // store from arg register
	    emitOp("movl",
		   regOperand(arg_location),
		   ptrDerefOperand(ptr_location));
	}
    }

    /** A helper that generates a pointer + offset load instruction of
     * data of the size of the given type */
    protected void emitLoadOffset(Location result_location, 
				  Location ptr_location, int offset,
				  ILType type)
		throws IOException {
	if (result_location == FPS) {
	    // load and push onto floating-point stack
	    emitOp("fldl", ptrOffsetDerefOperand(ptr_location, offset));
	} else {
	    // load into target register
	    emitOp("movl", 
		   ptrOffsetDerefOperand(ptr_location, offset), 
		   regOperand(result_location));
	}
    }

    /** A helper that generates a pointer + offset store instruction
     * of data of the size of the given type */
    protected void emitStoreOffset(Location ptr_location, int offset,
				   ILType type, Location arg_location)
		throws IOException {
	if (arg_location == FPS) {
	    // pop off floating-point stack and store
	    emitOp("fstpl", ptrOffsetDerefOperand(ptr_location, offset));
	} else {
	    // store from arg register
	    emitOp("movl",
		   regOperand(arg_location), 
		   ptrOffsetDerefOperand(ptr_location, offset));
	}
    }

    /** A helper that generates a pointer + offset + index load
     * instruction of data of the size of the given type */
    protected void emitLoadOffsetSequence(Location result_location, 
					  Location ptr_location, int offset, 
					  Location index_location, ILType type)
		throws IOException {
	if (result_location == FPS) {
	    // load and push onto floating-point stack
	    emitOp("fldl", 
		   ptrOffsetSequenceDerefOperand(ptr_location, offset,
						 index_location, type));
	} else {
	    // load into target register
	    emitOp("movl", 
		   ptrOffsetSequenceDerefOperand(ptr_location, offset,
						 index_location, type),
		   regOperand(result_location));
	}
    }

    /** A helper that generates a pointer + offset + index store
     * instruction of data of the size of the given type */
    protected void emitStoreOffsetSequence(Location ptr_location, int offset, 
					   Location index_location,
					   ILType type, 
					   Location arg_location)
		throws IOException {
	if (arg_location == FPS) {
	    // pop off floating-point stack and store
	    emitOp("fstpl",
		   ptrOffsetSequenceDerefOperand(ptr_location, offset,
						 index_location, type));
	} else {
	    // store from arg register
	    emitOp("movl",
		   regOperand(arg_location), 
		   ptrOffsetSequenceDerefOperand(ptr_location, offset,
						 index_location, type));
	}
    }

    /** A helper to evaluate and push the arguments to a call. return
     * the amount of stack space consumed by the arguments. */
    protected int emitCallArguments(List<ILExpr> args)
		throws IOException {
	// compute each argument and then push it on the stack, in
	// right-to-left order, so that the first argument is pushed
	// on last, and is at the lowest address, to match the stack
	// layout assumed by the callee.  also, compute the amount of
	// stack space allocated by all this pushing.
	int arg_stack_space = 0;
	ListIterator<ILExpr> iter = args.listIterator(args.size());
	while (iter.hasPrevious()) {
	    ILExpr arg = iter.previous();
	    
	    // evaluate the argument
	    Location arg_location = arg.codegen(this);

	    if (arg_location == FPS) {
		// the argument is a double -- we need to pop it from
		// the floating-point stack and push it on the call
		// stack

		// make space for a double on the data stack
		emitOp("subl", intOperand(8), regOperand(SP));

		// pop it off the floating-point stack and move it to
		// the new location on the data stack
		emitStore(SP, new ILDoubleType(), arg_location);

		// record amount of space allocated
		arg_stack_space += 8;
	    } else {
		// the argument is an int or pointer -- just push it

		// push it on the stack
		emitOp("pushl", regOperand(arg_location));

		// record amount of space allocated
		arg_stack_space += 4;
	    }

	    // deallocate the argument register
	    deallocateRegister(arg_location);
	}

	// return the amount of stack space consumed by these arguments
	return arg_stack_space;
    }

    /** A helper that returns a unique label each time it's called */
    protected String getNewLabel() {
	return ".B" + Integer.toString(label_number++);
    }

    /** Return the binary name for a library function with this C
     * source name */
    protected String runtimeFunName(String sourceName) {
	// see whether the runtime library functions require an
	// implicit underscore prefixed to their name (some C
	// compilers insert this automatically, and we have to match)
	if (runtimeFunsUnderscorePrefix) {
	    return "_" + sourceName;
	} else {
	    return sourceName;
	}
    }


    // Stack layout operations
    
    /** A helper that computes the layout of the current function's
     * stack frame */
    protected void layoutStackFrame(ILType return_type,
				    List<ILFormalVarDecl> formals, 
				    Map<String,ILVarDecl> decls) 
		throws IOException {
	// compute offsets of (non-formal) locals relative to the
	// first local (this is very similar to the calculations done
	// to lay out a record)
	int offset = 0; // we start at relative offset 0
	Iterator<Map.Entry<String,ILVarDecl>> decls_iter = 
	    decls.entrySet().iterator();
	while (decls_iter.hasNext()) {
	    Map.Entry<String,ILVarDecl> entry = decls_iter.next();
	    ILVarDecl var = entry.getValue();
	    if (! var.isFormal()) {
		// pad over the offset based on the necessary alignment
		int varAlign = var.getType().byteAlign(this);
		offset = Target.align(offset, varAlign);
		// remember its offset
		var.setRelativeByteOffset(offset);
		// add its size
		offset += var.getType().byteSize(this);
	    }
	}

	// compute offset where caller-save registers (all ints) will
	// go
	int callerSaveRegsOffset = Target.align(offset, 4);

	// reserve space for 4 of them, each 4 bytes long
	offset = callerSaveRegsOffset + 4*4;

	// the offset at this point is where the caller's frame
	// pointer is stored, and where this procedure's frame pointer
	// will point.  so the offset is also the amount we add to the
	// relative offset of each local when we access it.
	localsBase = -offset;

	// compute the offset relative to the frame pointer of the
	// caller-save registers area
	callerSaveRegsBase = callerSaveRegsOffset + localsBase;

	// the first formal is two pointers above where the frame
	// pointer points (skipping past the caller's saved frame
	// pointer and the saved return address)
	formalsBase = 2*4;

	// compute offsets of formals relative to the first formal
	offset = 0; // we start at offset 0
	Iterator<ILFormalVarDecl> formals_iter = formals.iterator();
	while (formals_iter.hasNext()) {
	    ILFormalVarDecl var = formals_iter.next();
	    // (formals aren't aligned, they're just pushed on
	    // densely, so don't do any alignment padding)
	    // remember its offset
	    var.setRelativeByteOffset(offset);
	    // add its size
	    offset += var.getType().byteSize(this);
	}

	if (commentOutput) {
	    // print out the whole stack layout info
	    emitComment("formals base offset: " + formalsBase);
	    emitComment("locals base offset: " + localsBase);
	    emitComment("caller-save regs offset: " + callerSaveRegsBase);
	    decls_iter = decls.entrySet().iterator();
	    while (decls_iter.hasNext()) {
		Map.Entry<String,ILVarDecl> entry = decls_iter.next();
		ILVarDecl var = entry.getValue();
		emitComment("offset of " + var.getName() + ": " + 
			    var.getByteOffset(this));
	    }
	}
    }

    /** return the base offset in the enclosing function's stack frame
     * for local variables */
    public int getLocalBaseOffset() {
	return localsBase;
    }

    /** return the base offset in the enclosing function's stack frame
     * for formal variables */
    public int getFormalBaseOffset() {
	return formalsBase;
    }


    // Register operations

    /** all the available registers */
    protected X86Register EAX = new X86Register("eax");
    protected X86Register EBX = new X86Register("ebx");
    protected X86Register ECX = new X86Register("ecx");
    protected X86Register EDX = new X86Register("edx");
    protected X86Register ESI = new X86Register("esi");
    protected X86Register EDI = new X86Register("edi");
    protected X86Register EBP = new X86Register("ebp");
    protected X86Register ESP = new X86Register("esp");

    /** the stack pointer */
    protected X86Register SP = ESP;

    /** the frame pointer */
    protected X86Register FP = EBP;

    /** two scratch registers */
    protected X86Register SCRATCH1 = ESI;
    protected X86Register SCRATCH2 = EDI;
    
    /** the set of allocatable registers */
    protected X86Register[] allocatableRegisters = {EAX, EBX, ECX, EDX};

    /** the floating-point stack location */
    protected X86FloatingPointStack FPS = new X86FloatingPointStack();

    /** a helper to initialize the register data structures */
    protected void initializeRegisters() {
	// mark the allocatable registers as such
	for (int i = 0; i < allocatableRegisters.length; i++) {
	    X86Register reg = allocatableRegisters[i];
	    reg.setAllocatable();
	}
    }

    /** A helper to allocate a free allocatable register (or the
     * floating-point stack, if a double) */
    protected Location allocateRegister(ILType type) {
	if (type.isIntType() || type.isPtrType()) {
	    // find the first available register
	    for (int i = 0; i < allocatableRegisters.length; i++) {
		X86Register reg = allocatableRegisters[i];
		if (! reg.isAllocated()) {
		    // found one free; allocate it
		    reg.allocate();
		    return reg;
		}
	    }
	    // didn't find any allocatable registers.  a real compiler
	    // would spill something to the stack here, but this isn't a
	    // real compiler...
	    throw new UnimplementedError(
			"too many registers needed; simplify expression");
	} else if (type.isDoubleType()) {
	    // allocate on the floating-point stack
	    return FPS;
	} else {
	    throw new InternalCompilerError(
		"unexpected type of data being allocated to a location");
	}
    }

    /** A helper to allocate a free allocatable register, preferring
     * the given register if available */
    protected Location allocateRegister(ILType type, 
					Location preference) {
	if (preference instanceof X86Register) {
	    if (type.isIntType() || type.isPtrType()) {
		X86Register pref_reg = (X86Register)preference;
		// see if the preferred register is available
		if (pref_reg.isAllocatable() && !pref_reg.isAllocated()) {
		    // it's free, so claim it
		    pref_reg.allocate();
		    return pref_reg;
		}
	    }
	}
	// preference couldn't be allocated, so just pick anything legal
	return allocateRegister(type);
    }

    /** A helper to deallocate a register */
    protected void deallocateRegister(Location location) {
	if (! (location instanceof X86Register)) {
	    // can't deallocate a non-register
	    return;
	}
	X86Register reg = (X86Register)location;
	reg.deallocate();
    }

    /** A helper to return whether the register has been allocated */
    protected boolean isAllocated(X86Register reg) {
	return reg.isAllocated();
    }

    /** A helper to save any live registers before a call */
    protected void saveRegisters() throws IOException {
	for (int i = 0; i < allocatableRegisters.length; i++) {
	    X86Register reg = allocatableRegisters[i];
	    if (reg.isAllocated()) {
		// need to save it

		// compute the offset of the saved place of this register
		int offset = callerSaveRegsBase + i*4;

		// treat it as an int, for saving purposes
		ILType type = new ILIntType();

		// generate the store
		emitStoreOffset(FP, offset, type, reg);
	    }
	}
    }

    /** A helper to restore any saved live registers after a call,
     * excluding the except location (which holds the call's result
     * and so was overwritten) */
    protected void restoreRegisters(Location except) throws IOException {
	for (int i = 0; i < allocatableRegisters.length; i++) {
	    X86Register reg = allocatableRegisters[i];
	    if (reg.isAllocated()) {
		if (reg == except) {
		    // not restored
		} else {
		    // need to restore it

		    // compute the offset of the saved place of this register
		    int offset = callerSaveRegsBase + i*4;

		    // treat it as an int, for saving purposes
		    ILType type = new ILIntType();

		    // generate the load
		    emitLoadOffset(reg, FP, offset, type);
		}
	    }
	}
    }

    /** A helper to return the location for the function's result,
     * based on its type */
    protected Location returnLocation(ILType return_type) {
	if (return_type.isDoubleType()) {
	    // double values are returned on the floating-point stack
	    return FPS;
	} else {
	    // integer & ptr values are returned in EAX
	    return EAX;
	}
    }


    // Type-specific size and alignment operations
    
    /** The size in bytes of an int value on this target */
    public int intByteSize() {
	return 4;
    }

    /** The alignment constraint in bytes of an int value on this
     * target */
    public int intByteAlign() {
	return 4;
    }
    
    /** The size in bytes of an int value on this target */
    public int doubleByteSize() {
	return 8;
    }

    /** The alignment constraint in bytes of an int value on this
     * target */
    public int doubleByteAlign() {
	return 8;
    }
    
    /** The size in bytes of an int value on this target */
    public int ptrByteSize() {
	return 4;
    }

    /** The alignment constraint in bytes of an int value on this
     * target */
    public int ptrByteAlign() {
	return 4;
    }
}
