/** The class for the environment for a class, which stores
 * information about each of the class's methods. */

package Evaluator;

import AST.*;
import Typechecker.*;
import Helpers.*;
import java.util.*;

public class ClassEnvironment extends NestedEnvironment {
    // Instance variables

    /** the name of the class that this is an environment for */
    protected String class_name;

    /** the environment of the class's superclass, or null if the
     * class doesn't extend any other class */
    protected ClassEnvironment superclass_env;

    /** the methods declared in this class */
    protected Map<String,MethodDecl> methods;


    // Constructor

    /** create the environment for a class, given the environment for
     * the scope enclosing the class declaration (which is the global
     * environment) */
    public ClassEnvironment(String cn, Environment enclosing_env) {
	super(enclosing_env);
	class_name = cn;
	superclass_env = null;
	methods = new HashMap<String,MethodDecl>();
    }


    // Accessors

    /** return the name of the class that this is the environment
     * for */
    public String getClassName() {
	return class_name;
    }

    /** record the class's superclass's environmemnt, if the class has a
     * superclass */
    public void setSuperclassEnvironment(ClassEnvironment se) {
	superclass_env = se;
    }

    /** look up the current contents of a local variable */
    public Value lookupLocalVariable(String name, int line_number)
		throws EvalCompilerException {
	// no longer in a code environment, so no more local variables
	// to find
	throw new InternalCompilerError(
		    "referencing undefined local variable \"" + name + "\"",
		    line_number);
    }

    /** assign to a local variable */
    public void assignLocalVariable(String name, Value value, 
				    int line_number) {
	// no longer in a code environment, so no more local variables
	// to find
	throw new InternalCompilerError(
		    "assigning to undefined local variable \"" + name + "\"",
		    line_number);
    }

    /** declare a new method in this class */
    public void declareMethod(String name, MethodDecl md, int line_number) {
	if (methods.containsKey(name)) {
	    // we've already declared a method with this same name
	    throw new InternalCompilerError(
		    "redeclaring method named \"" + name + "\"", line_number);
	} else {
	    // remember this method
	    methods.put(name, md);
	}
    }
    
    /** look up the method with the given name in this class or one of
     * its superclasses.  return its declaration and the environment
     * for the class in which it was found. */
    public MethodDeclAndEnv lookupMethod(String name, int line_number) {
	MethodDecl md = methods.get(name);
	if (md == null) {
	    // method not found in this class
	    if (superclass_env == null) {
		// and no superclasses to search, so this is a method
		// lookup error
		throw new InternalCompilerError(
		    "calling undefined method \"" + name + "\"", line_number);
	    } else {
		// recursively search the superclass
		return superclass_env.lookupMethod(name, line_number);
	    }
	} else {
	    // method found
	    // return the method along with its enclosing class env
	    return new MethodDeclAndEnv(md, this);
	}
    }
}
