/** The class for the environment for the body of a method or of a
 * nested block statement. */

package Evaluator;

import Helpers.*;
import java.util.*;

public class CodeEnvironment extends NestedEnvironment {
    // Instance variables

    /** the values of local & formal variables in this scope */
    protected Map<String,Value> local_variables;


    // Constructor

    /** create the environment for a statement list, given the
     * environment of its lexically enclosing scope */
    public CodeEnvironment(Environment enclosing_env) {
	super(enclosing_env);
	local_variables = new HashMap<String,Value>();
    }


    // Accessors

    /** declare and initialize a local variable, checking for
     * duplicates */
    public void declareLocalVariable(String name, Value value, 
				     int line_number) {
	if (local_variables.containsKey(name)) {
	    // we've already declared a variable with this same name
	    throw new InternalCompilerError(
			"redeclaring variable named \"" + name + "\"", 
			line_number);
	} else {
	    // declare this variable, initializing it with the given
	    // value
	    local_variables.put(name, value);
	}
    }

    /** look up the current contents of a local variable */
    public Value lookupLocalVariable(String name, int line_number)
		throws EvalCompilerException {
	Value value = local_variables.get(name);
	if (value == null) {
	    // variable not declared in this scope; search its
	    // enclosing scope
	    return enclosing_env.lookupLocalVariable(name, line_number);
	} else {
	    // it's declared here; return its contents
	    return value;
	}
    }

    /** assign to a local variable */
    public void assignLocalVariable(String name, Value value, 
				    int line_number) {
	if (local_variables.containsKey(name)) {
	    // assign to this variable
	    local_variables.put(name, value);
	} else {
	    // variable not declared in this scope; search its
	    // enclosing scope
	    enclosing_env.assignLocalVariable(name, value, line_number);
	}
    }


    // Tracing operations

    /** return the number of callers of this method environment (0 =
     * the main class's main method) */
    public int caller_depth() {
	return enclosing_env.caller_depth();
    }

    /** return a printable string prefix to use for tracing
     * information about this environment */
    public String caller_depth_indent() {
	// (would be more efficient using StringBuffers)
	String s = "";
	for (int i = caller_depth(); i > 0; i--) {
	    s += " ";
	}
	return s;
    }
}
