/** This is the class for IL (intermediate language) function
 * declarations */

package IL;

import Typechecker.*;
import CodeGen.*;
import Helpers.*;
import java.io.*;
import java.util.*;

public class ILFunDecl {
    // Instance variables

    // the main ones, used to generate target code after the function
    // is fully constructed
    String fun_name;     // the name of the function (which should be
			 // unique among other functions and global
			 // variables)
    ILType return_type;  // the type of its result
    List<ILFormalVarDecl> formals;  // in left-to-right order
    Map<String,ILVarDecl> decls;    // locals + formals + temps
    List<ILStmt> stmts;  // in execution order; control should not fall off end

    // some helpers used only during intermediate-code generation,
    // a.k.a. lowering
    ILProgram program;  // the program containing this function
    int var_counter = 0;
    int label_counter = 0;


    // Constructor

    // create a new fun, initially without any formals, locals, or
    // statements, in IL program p with return type rt and name n
    public ILFunDecl(ILProgram p, ILType rt, String n) { 
	program = p;
	return_type = rt;
	fun_name = n;
	formals = new LinkedList<ILFormalVarDecl>();
	decls = new HashMap<String,ILVarDecl>();
	stmts = new LinkedList<ILStmt>();
    }


    // Basic operations to augment the function with parts not
    // provided during construction

    // declare a new local variable (callable in any order)
    public void declareLocal(LocalVarInterface vi) {
	ILLocalVarDecl decl = 
	    new ILLocalVarDecl(vi.getType().lower(), varName(vi));
	declareVar(decl);
    }

    // declare a new formal (should be called for each formal, in
    // left-to-right order)
    public void declareFormal(LocalVarInterface vi) {
	ILFormalVarDecl decl = 
	    new ILFormalVarDecl(vi.getType().lower(), varName(vi));
	// formals are recorded two ways: one in order, and one indexed by name
	formals.add(decl);
	declareVar(decl);
    }

    // a helper to declare a new local or formal variable
    protected void declareVar(ILVarDecl decl) {
	String name = decl.getName();
	if (decls.containsKey(name)) {
	    throw new InternalCompilerError(
			"redeclaring local " + name + " in " + fun_name);
	} else {
	    decls.put(name, decl);
	}
    }

    // append a new statement to the list of statements
    public void addStmt(ILStmt stmt) {
	stmts.add(stmt);
    }

    // add a new global var decl to the program containing this function
    public void declareGlobalVar(ILGlobalVarDecl decl) {
	program.addGlobalVarDecl(decl);
    }


    // Accessors

    // return an ILVar for a reference to the given local variable or
    // formal declared in this function
    public ILVar lookupVar(LocalVarInterface vi) {
	String name = varName(vi);
	ILVarDecl decl = decls.get(name);
	if (decl == null) {
	    throw new InternalCompilerError(
		"referencing undefined local " + name + 
		" in " + fun_name);
	} else {
	    return new ILVar(decl);
	}
    }

    // return an ILVar for a reference to the 'this' formal parameter
    // declared in this function
    public ILVar lookupThis() {
	ILVarDecl decl = decls.get("this");
	if (decl == null) {
	    throw new InternalCompilerError(
		"referencing undefined local 'this' in " + fun_name);
	} else {
	    return new ILVar(decl);
	}
    }

    // return an ILGlobalVar for a reference to the given global variable
    public ILGlobalVar lookupGlobalVar(String name) {
	return program.lookupGlobalVar(name);
    }


    // To allow for distinguishing different local variables with the
    // same name but in different (source) scopes without having
    // scopes in the IL functions, we construct the IL variable's name
    // based on both the source variable's name and the unique ID of
    // its source declaration.  Except that we allow declarations of
    // 'this' to go through unchanged, so that we can reference 'this'
    // directly.
    protected String varName(LocalVarInterface decl) {
	String declName = decl.getName();
	if (declName.equals("this")) {
	    return "this";
	} else {
	    return decl.getName() + "." + decl.hashCode();
	}
    }


    // Operations for allocating fresh temporary variables and labels

    // declare a fresh, new temp variable of the given type, and
    // return an ILVar referring to it
    public ILVar newTempVar(ILType type) {
	// create a new temp var whose name is derived from a counter
	// that is incremented each time a new temp is needed
	String name = ".T." + Integer.toString(var_counter++);
	// create its declaration
	ILVarDecl decl = new ILLocalVarDecl(type, name);
	// record it and its type
	declareVar(decl);
	// return a reference to it
	return new ILVar(decl);
    }

    // return a fresh, new label
    public ILLabel newLabel() {
	// create a new label whose name is derived from a counter
	// that is incremented each time a new label is needed, plus
	// the name of the enclosing function (to make it globally
	// unique in the output file)
	return new ILLabel("L" + label_counter++ + "." + fun_name);
    }

    // declare a fresh, new global variable of the given type,
    // initialized to the given value, and return an ILGlobalVar
    // referring to it
    public ILGlobalVar newInitializedGlobalVar(ILType type, 
					       ILValue initValue) {
	return program.newInitializedGlobalVar(type, initValue);
    }


    // Printing
    public String toString() {
	String s = return_type.toString() + " " + fun_name + "(";
	Iterator<ILFormalVarDecl> formals_iter = formals.iterator();
	boolean first = true;
	while (formals_iter.hasNext()) {
	    ILFormalVarDecl formal = formals_iter.next();
	    if (first) {
		first = false;
	    } else {
		s += ", ";
	    }
	    s += formal.toString();
	}
	s += ") {\n";
	Iterator<Map.Entry<String,ILVarDecl>> entries_iter =
	    decls.entrySet().iterator();
	while (entries_iter.hasNext()) {
	    Map.Entry<String,ILVarDecl> entry = entries_iter.next();
	    ILVarDecl var = entry.getValue();
	    if (! var.isFormal()) {
		s += "  " + var.toString() + ";\n";
	    }
	}
	Iterator<ILStmt> stmts_iter = stmts.iterator();
	while (stmts_iter.hasNext()) {
	    ILStmt stmt = stmts_iter.next();
	    s += "  " + stmt.toString() + "\n";
	}
	s += "}";
	return s;
    }


    // Target code generation operations

    /** generate code for this function */
    public void codegen(Target target) throws IOException {
	// lay out the function's stack frame, generate whatever the
	// target needs to declare the function, and set up its stack
	// frame
	target.emitFunctionPrologue(return_type, fun_name, formals, decls);

	// generate code for each statement
	Iterator<ILStmt> iter = stmts.iterator();
	while (iter.hasNext()) {
	    ILStmt stmt = iter.next();
	    stmt.codegen(target);
	}

	// generate whatever the target needs at the end of the function
	target.emitFunctionEpilogue(fun_name);
    }
}
