/** This is the class for IL (intermediate language) programs.  A
 * program has an unordered list of global variables and functions. */

package IL;

import CodeGen.*;
import Helpers.*;
import java.io.*;
import java.util.*;

public class ILProgram {
    // Instance variables

    // the main ones, used to generate target code after the program
    // is fully constructed

    /** the global variable declarations in the program */
    protected List<ILGlobalVarDecl> vars;

    /** the function declarations in the program */
    protected List<ILFunDecl> funs;

    // some helpers used only during intermediate-code generation,
    // a.k.a. lowering
    int global_var_counter = 0;


    // Constructor

    /** create a new, empty IL program */
    public ILProgram() { 
	vars = new LinkedList<ILGlobalVarDecl>();
	funs = new LinkedList<ILFunDecl>();
    }


    // Adding global declarations

    // add a new global var decl
    public void addGlobalVarDecl(ILGlobalVarDecl decl) {
	vars.add(decl);
    }

    // add a new global fun decl
    public void addFunDecl(ILFunDecl fun) {
	funs.add(fun);
    }


    // Accessors

    // return an ILGlobalVar for a reference to the given global
    // variable (or function)
    public ILGlobalVar lookupGlobalVar(String name) {
	// global vars can be referenced before they're declared (for
	// one, to allow mutual recursion across global variables &
	// functions), so just return a reference without checking for
	// whether it's been declared
	return new ILGlobalVar(name);
    }


    // Helper operations for allocating fresh global vars

    // return a fresh, new global variable name of the given type,
    // whose declaration is initialized to the given value
    public ILGlobalVar newInitializedGlobalVar(ILType type, 
					       ILValue initValue) {
	// create a new global var whose name is derived from a counter
	// that is incremented each time a new global is needed
	String name = "G." + Integer.toString(global_var_counter++);
	// create its declaration
	ILGlobalVarDecl decl =
	    new ILInitializedGlobalVarDecl(type, name, initValue);
	// record it
	addGlobalVarDecl(decl);
	// return a reference to it
	return new ILGlobalVar(name);
    }


    // Printing
    public String toString() {
	String s = "";
	Iterator<ILGlobalVarDecl> vars_iter = vars.iterator();
	while (vars_iter.hasNext()) {
	    ILGlobalVarDecl var = vars_iter.next();
	    s += var.toString() + "\n";
	}
	Iterator<ILFunDecl> funs_iter = funs.iterator();
	while (funs_iter.hasNext()) {
	    ILFunDecl fun = funs_iter.next();
	    s += fun.toString() + "\n";
	}
	return s;
    }


    // Target code generation operations

    /** generate code for this program */
    public void codegen(Target target) throws IOException {
	// generate whatever the target needs at the start of the output file
	target.emitProgramPrologue();

	// generate all the global variable declarations
	Iterator<ILGlobalVarDecl> vars_iter = vars.iterator();
	while (vars_iter.hasNext()) {
	    ILGlobalVarDecl var = vars_iter.next();
	    var.codegen(target);
	}

	// generate all the function declarations
	Iterator<ILFunDecl> funs_iter = funs.iterator();
	while (funs_iter.hasNext()) {
	    ILFunDecl fun = funs_iter.next();
	    fun.codegen(target);
	}

	// generate whatever the target needs at the end of the output file
	target.emitProgramEpilogue();
    }
}
