/** The class for IL sequence types.  A value of ILSequenceType is a
 * sequence of zero or more values of the sequence's element type; the
 * length of this sequence is not statically known nor does the IL
 * language automatically track the length dynamically.  Since values
 * of ILSequenceType don't have statically known length, an
 * ILSequenceType can only appear as the target of a pointer or as the
 * last member of a record that's the target of a pointer. */

package IL;

import CodeGen.*;
import java.util.*;

public class ILSequenceType extends ILType {
    // Instance variable
    protected ILType elem_type;    // the sequence's element type


    // Constructor
    public ILSequenceType(ILType et) {
	elem_type = et;
    }


    // Accessors

    // if a sequence type, then return the type of the sequence elements.
    // otherwise report an internal error
    public ILType getSequenceElemType() {
	return elem_type;
    }


    // Printing operations

    // return a printable string representation of this type.
    public String toString() { 
	return "seq<" + elem_type.toString() + ">"; 
    }


    // Target code generation operations

    /** compute the layout of this type */
    public void doLayout(Target target) {
	// lay out the element type
	elem_type.doLayout(target);
    }

    /** return the size in bytes of a value of this type on the given
     * target machine */
    public int byteSize(Target target) {
	// for a sequence, we return the fixed size, ignoring any
	// elements, which is always 0 (when we allocate a real sequence,
	// we'll add in the space required for the number of elements
	// allocated)
	return 0;
    }

    /** return the size in bytes of each sequence element, assuming this
     * type ends with a sequence type.  report an error if not. */
    public int sequenceElementByteSize(Target target) {
	return elem_type.byteSize(target);
    }

    /** return the alignment constraint in bytes of a value of this
     * type on the given target machine, i.e., at what multiple of
     * bytes must a value of this type begin on */
    public int byteAlign(Target target) {
	// the alignment constraint of a sequence is that of its element
	// type
	return elem_type.byteAlign(target);
    }
}
