/** This is the main class for the MiniJava compiler */

import Scanner.*;
import Parser.*;
import AST.*;
import Typechecker.*;
import Evaluator.*;
import IL.*;
import CodeGen.*;
import CodeGen.X86.*;
import CodeGen.MIPS.*;
import Helpers.*;
import java_cup.runtime.Symbol;
import java.io.*;

public class MiniJava {
    boolean verbose = false; // print notices about each phase

    boolean justHelp = false; // just print the help message?
    boolean justScan = false; // stop after scanning?
    boolean justParse = false; // stop after parsing?
    boolean justTypecheck = false; // stop after typechecking?
    boolean justEvaluate = false; // do evaluation, and then stop?
    boolean justLower = false; // stop after lowering to intermediate code?

    boolean printTokens = false; // print each token scanned?
    boolean printAST = false; // print AST after parsing?
    boolean printSymbolTables = false; // print sym.tabs. after typechecking?
    boolean printCalls = false; // print calls & returns during evaluation?
    boolean printIL = false; // print IL after lowering?
    boolean printCode = false; // print generated code after codegen?
    
    boolean optimize = false; // enable optimizations?

    boolean runtimeFunsUnderscorePrefix = false; // runtime library
						 // functions compiled
						 // with an implicit
						 // leading
						 // underscore?

    Target target = new X86Target(); // the target machine to compile for

    String file_name; // the MiniJava source file to compile

    /** This is the main entry point for invoking the MiniJava compiler */
    public static void main(String[] args){
	try {
	    MiniJava c = new MiniJava();
	    c.processArguments(args);
	    c.compile();
	    if (c.verbose) {
		// report successful completion
		System.out.println("** Done.");
	    }
	} catch (CompilerException e) {
	    // an error in the user's arguments or input, or some
	    // other kind of error that we've already handled in the
	    // appropriate way (not a bug that the error got here)
	    System.err.println(e.getMessage());
	} catch (Throwable e) {
	    // yuck: some kind of error in the compiler implementation
	    // that we're not expecting (a bug!)
	    System.err.println("Unexpected internal compiler error: " + 
			       e.toString());
	    // print out a stack dump
	    e.printStackTrace();
	}
    }

    /** Create a compiler instance */
    MiniJava() { }

    /** A helper to process the command-line arguments */
    void processArguments(String[] args) {
	int num_args = args.length;
	if (num_args == 0) {
	    // no arguments => give a help message
	    System.err.println("expecting (at least) a filename argument");
	    justHelp = true;
	} else {
	    // process options
	    for (int i = 0; i < num_args-1; i++) {
		String option = args[i];
		processOption(option);
	    }
	    String last_arg = args[num_args-1];
	    if (last_arg.equals("-help")) {
		// OK to skip a filename if we have just -help
		justHelp = true;
	    } else if (last_arg.charAt(0) == '-') {
		// oops: finishing with an option, not a file name
		processOption(last_arg);
		System.err.println("expecting a filename argument");
		justHelp = true;
	    } else {
		// the last argument is a fine-looking filename
		file_name = last_arg;
	    }
	}
    }

    /** A helper to process the command-line arguments */
    void processOption(String option) {
	if (option.equals("-verbose")) {
	    verbose = true;
	} else if (option.equals("-x86")) {
	    // the default; nothing to set
	} else if (option.equals("-mips")) {
	    System.out.println("sorry, MIPS target not implemented yet");
	    // target = new MIPSTarget();
	} else if (option.equals("-help")) {
	    justHelp = true;
	} else if (option.equals("-scan")) {
	    justScan = true;
	} else if (option.equals("-parse")) {
	    justParse = true;
	} else if (option.equals("-typecheck")) {
	    justTypecheck = true;
	} else if (option.equals("-evaluate")) {
	    justEvaluate = true;
	} else if (option.equals("-lower")) {
	    justLower = true;
	} else if (option.equals("-codegen")) {
	    // the default; nothing to set
	} else if (option.equals("-optimize")) {
	    optimize = true;
	} else if (option.equals("-printTokens")) {
	    printTokens = true;
	} else if (option.equals("-printAST")) {
	    printAST = true;
	} else if (option.equals("-printSymbolTables")) {
	    printSymbolTables = true;
	} else if (option.equals("-printCalls")) {
	    printCalls = true;
	} else if (option.equals("-printIL")) {
	    printIL = true;
	} else if (option.equals("-printCode")) {
	    printCode = true;
	} else if (option.equals("-runtimeFunsUnderscorePrefix")) {
	    runtimeFunsUnderscorePrefix = true;
	} else {
	    // unexpected option; print a help message and exit
	    System.err.println("unexpected option: " + option);
	    justHelp = true;
	}
    }

    void help() {
	String helpMessage = "\n" +
	    "MiniJava [option...] file.java\n" +
	    "where option can be:\n" +
	    "  -x86:        generate code for the x86 (the default)\n" +
	 // "  -mips:       generate code for the MIPS\n" +
	    "\n" +
	    "  -scan:       stop after scanning\n" +
	    "  -parse:      stop after parsing\n" +
	    "  -typecheck:  stop after typechecking\n" +
	    "  -evaluate:   evaluate program directly\n" +
	    "  -lower:      stop after lowering\n" +
	    "  -codegen:    do all passes (the default)\n" +
	    "\n" +
	    "  -optimize:   enable optimizations\n" +
	    "\n" +
	    "  -printTokens: print tokens as they're scanned\n" +
	    "  -printAST:    print parsed AST after it's parsed\n" +
	    "  -printSymbolTables: print symbol tables after typechecking\n" +
	    "  -printCalls:  print calls and returns during evaluation\n" +
	    "  -printIL:     print intermediate language after lowering\n" +
	    "  -printCode:   print generated code after codegen\n" +
	    "\n" +
	    "  -verbose:     print progress notices\n" +
	    "\n" +
	    "  -runtimeFunsUnderscorePrefix: runtime library functions compiled with an\n" +
	    "                                implicit underscore prefix, e.g. on\n" +
	    "                                Windows/Cygwin, but not Linux\n" +
	    "\n" +
	    "  -help:        print this message\n";
	System.out.println(helpMessage);
    }

    /** A helper to do the work */
    void compile() throws CompilerException {
	if (justHelp) {
	    help();
	    return;
	}

	if (verbose) {
	    System.out.println("** Opening input file \"" + file_name + 
			       "\"...");
	}

	// open the input file
	FileReader input_file;
	try {
	    input_file = new FileReader(file_name);
	} catch (FileNotFoundException e) {
	    throw new CompilerException("input file \"" + file_name + 
					"\" not found");
	}

	// create a scanner on the input file
	scanner s = new scanner(input_file);

	if (justScan) {
	    // scan the input

	    if (verbose) {
		System.out.println("** Scanning...");
	    }

	    try {
		Symbol t = s.next_token();
		while (t.sym != sym.EOF){ 
		    if (printTokens) {
			// print each token that we scan
			System.out.print(s.symbolToString(t) + " ");
		    }
		    t = s.next_token(); 
		}
		if (printTokens) {
		    System.out.print("\n");
		}
	    } catch (IOException e) {
		throw new CompilerException("Input file read error: " + 
					    e.toString());
	    }
	    return;
	}

	// scan and parse the input

	if (verbose) {
	    System.out.println("** Scanning and Parsing...");
	}

	if (printTokens) {
	    System.out.println("sorry, cannot -printTokens unless -scan");
	}

	// create a parser on the scanner
	parser p = new parser(s);

	Symbol root;
	try {
	    root = p.parse();
	} catch (CompilerException e) {
	    // an expected kind of error from the parser
	    throw e;
	} catch (Exception e) {
	    // an unexpected kind of error; a bug!
	    throw new InternalCompilerError(e.toString());
	}
	Program program = (Program)root.value;

	if (printAST) {
	    // pretty-print the resulting parse tree
	    System.out.println(program.toString());
	}

	if (justParse) {
	    return;
	}

	// typecheck the input

	if (verbose) {
	    System.out.println("** Typechecking...");
	}

	GlobalSymbolTable st = program.typecheck();

	if (printSymbolTables) {
	    System.out.println(st.toString());
	}

	if (justTypecheck) {
	    return;
	}

	if (justEvaluate) {
	    // do evaluation instead of code generation

	    if (verbose) {
		System.out.println("** Evaluating...");
	    }

	    program.evaluate(printCalls);

	    return;
	}

	// lower the input program (i.e. translate it into
	// intermediate language)

	if (verbose) {
	    System.out.println("** Lowering...");
	}

	ILProgram il_program = program.lower();

	if (printIL) {
	    System.out.println(il_program.toString());
	}

	if (justLower) {
	    return;
	}

	// generate target code from the lowered program

	if (verbose) {
	    System.out.println("** Generating target code...");
	}

	try {

	    if (optimize) {
		// turn on optimizations
		target.enableOptimizations();
	    }

	    // initialize target, creating output file
	    target.initialize(file_name, printCode, 
			      runtimeFunsUnderscorePrefix);

	    // generate code from the IL program, printing out any emitted
	    // assembly if desired
	    il_program.codegen(target);

	    // finish up the output file generation
	    target.finalize();

	} catch (IOException e) {
	    throw new CompilerException("Output file write error: " + 
					e.toString());
	}

	// done!
    }
}
