/* JFlex specification for the lexical analyzer for MiniJava */

package Scanner;

import java_cup.runtime.*;
import Parser.sym;
import Helpers.*;

%%

%public
%final
%class scanner
%yylexthrow CompilerException
%unicode
%cup
%line
%column

%{
  // note that these Symbol constructors are abusing the Symbol
  // interface to use Symbol's left and right fields as line and column
  // fields instead
  private Symbol symbol(int type) {
    return new Symbol(type, yyline+1, yycolumn+1);
  }
  private Symbol symbol(int type, Object value) {
    return new Symbol(type, yyline+1, yycolumn+1, value);
  }

  // print out a symbol (aka token) nicely
  public String symbolToString(Symbol s) {
    switch (s.sym) {
      case sym.CLASS: return "class";
      case sym.PUBLIC: return "public";
      case sym.STATIC: return "static";
      case sym.EXTENDS: return "extends";
      case sym.VOID: return "void";
      case sym.INT: return "int";
      case sym.BOOLEAN: return "boolean";
      case sym.IF: return "if";
      case sym.ELSE: return "else";
      case sym.WHILE: return "while";
      case sym.RETURN: return "return";
      case sym.NULL: return "null";
      case sym.TRUE: return "true";
      case sym.FALSE: return "false";
      case sym.THIS: return "this";
      case sym.NEW: return "new";
      case sym.STRING: return "String";
      case sym.MAIN: return "main";
      case sym.SYSTEM_OUT_PRINTLN: return "System.out.println";
      case sym.AND_AND: return "&&";
      case sym.LESSEQUAL: return "<=";
      case sym.LESSTHAN: return "<";
      case sym.GREATEREQUAL: return ">=";
      case sym.GREATERTHAN: return ">";
      case sym.EQUALS_EQUALS: return "==";
      case sym.EXCLAIM_EQUALS: return "!=";
      case sym.PLUS: return "+";
      case sym.MINUS: return "-";
      case sym.STAR: return "*";
      case sym.SLASH: return "/";
      case sym.EXCLAIM: return "!";
      case sym.OPEN_PAREN: return "(";
      case sym.CLOSE_PAREN: return ")";
      case sym.OPEN_BRACE: return "{";
      case sym.CLOSE_BRACE: return "}";
      case sym.OPEN_BRACKET: return "[";
      case sym.CLOSE_BRACKET: return "]";
      case sym.EQUALS: return "=";
      case sym.SEMICOLON: return ";";
      case sym.COMMA: return ",";
      case sym.PERIOD: return ".";
      case sym.IDENTIFIER: return (String)s.value;
      case sym.INT_LITERAL: return ((Integer)s.value).toString();
      case sym.EOF: return "<EOF>";
      case sym.error: return "<ERROR>";
      default: return "<UNEXPECTED TOKEN " + s.toString() + ">";
    }
  }
%}

/* Helper definitions */
letter = [a-zA-Z]
digit = [0-9]
eol = [\r\n]
white = {eol}|[ \t]

%%

/* Token definitions */

/* reserved words */
/* (put here so that reserved words take precedence over identifiers) */
"class" { return symbol(sym.CLASS); }
"public" { return symbol(sym.PUBLIC); }
"static" { return symbol(sym.STATIC); }
"extends" { return symbol(sym.EXTENDS); }
"void" { return symbol(sym.VOID); }
"int" { return symbol(sym.INT); }
"boolean" { return symbol(sym.BOOLEAN); }
"if" { return symbol(sym.IF); }
"else" { return symbol(sym.ELSE); }
"while" { return symbol(sym.WHILE); }
"return" { return symbol(sym.RETURN); }
"null" { return symbol(sym.NULL); }
"true" { return symbol(sym.TRUE); }
"false" { return symbol(sym.FALSE); }
"this" { return symbol(sym.THIS); }
"new" { return symbol(sym.NEW); }
/* the following aren't normally reserved words, but are reserved so
   they can be handled more easily when compiling MiniJava: */
"String" { return symbol(sym.STRING); }
"main" { return symbol(sym.MAIN); }
"System.out.println" { return symbol(sym.SYSTEM_OUT_PRINTLN); }

/* operators */
"+" { return symbol(sym.PLUS); }
"-" { return symbol(sym.MINUS); }
"*" { return symbol(sym.STAR); }
"/" { return symbol(sym.SLASH); }
"!" { return symbol(sym.EXCLAIM); }
"<" { return symbol(sym.LESSTHAN); }
"<=" { return symbol(sym.LESSEQUAL); }
">=" { return symbol(sym.GREATEREQUAL); }
">" { return symbol(sym.GREATERTHAN); }
"==" { return symbol(sym.EQUALS_EQUALS); }
"!=" { return symbol(sym.EXCLAIM_EQUALS); }
"&&" { return symbol(sym.AND_AND); }

/* delimiters */
"(" { return symbol(sym.OPEN_PAREN); }
")" { return symbol(sym.CLOSE_PAREN); }
"{" { return symbol(sym.OPEN_BRACE); }
"}" { return symbol(sym.CLOSE_BRACE); }
"[" { return symbol(sym.OPEN_BRACKET); }
"]" { return symbol(sym.CLOSE_BRACKET); }
"=" { return symbol(sym.EQUALS); }
";" { return symbol(sym.SEMICOLON); }
"," { return symbol(sym.COMMA); }
"." { return symbol(sym.PERIOD); }

/* identifiers */
{letter} ({letter}|{digit})* { return symbol(sym.IDENTIFIER, yytext()); }

/* integers */
{digit}+ {
    String number = yytext();
    try {
        return symbol(sym.INT_LITERAL, Integer.valueOf(number));
    } catch (NumberFormatException e) { 
	throw new LexicalCompilerException(
		"integer literal out of range: " + number,
		yyline+1, yycolumn+1);
    } }


/* whitespace */
{white}+ { /* ignore whitespace */ }

/* lexical errors (put last so other matches take precedence) */
. { throw new LexicalCompilerException(
	"unexpected character in input: '" + yytext() + "'", 
	yyline+1, yycolumn+1);
  }
