/** The class for the symbol table for a class, which stores the
 * interfaces of its methods and class variables. */

package Typechecker;

import AST.*;
import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class ClassSymbolTable extends NestedSymbolTable {
    // Instance variables

    /** the name of the class for which this is the symbol table */
    protected String class_name;

    /** the symbol table of the class's superclass, or null if the
     * class doesn't extend any other class */
    protected ClassSymbolTable superclass_st;

    /** the interface of class variables declared in this scope */
    protected Map<String,ClassVarInterface> class_variables;

    /** the (non-static) methods declared in this class */
    protected Map<String,MethodInterface> methods;

    /** the IL record for the layout of the vtbl for this class */
    protected ILRecordType vtbl_type;

    /** the IL record for the layout of instances of this class */
    protected ILRecordType instance_type;


    // Constructor

    /** create the symbol table for a class, given the symbol table
     * for the scope enclosing the class declaration (which is the
     * global symbol table for top-level class declarations) */
    public ClassSymbolTable(String cn, SymbolTable enclosing_symbol_table) {
	super(enclosing_symbol_table);
	class_name = cn;
	superclass_st = null;
	class_variables = new HashMap<String,ClassVarInterface>();
	methods = new HashMap<String,MethodInterface>();
	vtbl_type = null;
	instance_type = null;
    }


    // Accessors

    /** return the class's name */
    public String getClassName() {
	return class_name;
    }

    /** record the class's superclass's symbol table, if the class has
     * a superclass */
    public void setSuperclass(ClassSymbolTable sst) {
	superclass_st = sst;
    }

    /** return the resolved type corresponding to this class */
    public ResolvedType getType() {
	return new ResolvedClassType(this);
    }


    // Printing

    /** Return a printable string representation of the symbol table
     * information */
    public String toString(String indent) {
	String s = "ClassSymbolTable{\n";
	String nested_indent = ASTNode.nested_indent(indent);
	if (superclass_st != null) {
	    s += nested_indent + "extends " + 
		superclass_st.getClassName() + ";\n";
	}
	Iterator<Map.Entry<String,ClassVarInterface>> class_vars_iter =
	    class_variables.entrySet().iterator();
	while (class_vars_iter.hasNext()) {
	    Map.Entry<String,ClassVarInterface> entry = class_vars_iter.next();
	    String var_name = entry.getKey();
	    ClassVarInterface var_interface = entry.getValue();
	    s += indent + var_name + ": " + var_interface.toString() + "\n";
	}
	Iterator<Map.Entry<String,MethodInterface>> methods_iter =
	    methods.entrySet().iterator();
	while (methods_iter.hasNext()) {
	    Map.Entry<String,MethodInterface> entry = methods_iter.next();
	    String method_name = entry.getKey();
	    MethodInterface method_interface = entry.getValue();
	    s += nested_indent + method_name + ": " + 
		method_interface.toString() + "\n";
	}
	s += indent + "}";
	return s;
    }


    // Typechecking operations

    /** Return whether the receiver is a direct or indirect subclass
     * of the argument */
    public boolean inheritsFrom(ClassSymbolTable other) {
	if (superclass_st == null) {
	    // no superclasses to search
	    return false;
	} else if (superclass_st == other) {
	    // a direct subclass
	    return true;
	} else {
	    // search recursively, to see if it's an indirect subclass
	    return superclass_st.inheritsFrom(other);
	}
    }

    /** declare a new method in this class, checking for duplicates */
    public void declareMethod(String name, MethodInterface mi,
			      int line_number)
		throws TypecheckCompilerException {
	if (methods.containsKey(name)) {
	    // we've already declared a method with this same name,
	    // and static overloading isn't supported by MiniJava
	    throw new TypecheckCompilerException(
				"redeclaring method named \"" + name + "\"",
				line_number);
	} else {
	    // check that this method doesn't overload improperly with
	    // any inherited method
	    if (superclass_st != null) {
		MethodInterface inherited_mi = 
		    superclass_st.lookupMethodOrNull(name);
		if (inherited_mi != null) {
		    // we're overloading an inherited method.  OK if
		    // they have the same argument types, i.e., if
		    // we're overriding the inherited method
		    if (mi.overrides(inherited_mi)) {
			// they're the same argument and result types, so OK.
			// remember that this method overrides the other.
			mi.setRootMethod(inherited_mi.getRootMethod());
		    } else {
			// they overload, not override, which isn't
			// supported by MiniJava
			throw new TypecheckCompilerException(
				"redeclaring inherited method named \"" +
				name + "\"", line_number);
		    }
		} else {
		    // there was no inherited method, so we're OK
		    mi.setRootMethod(mi);
		}
	    } else {
		// no superclass, so no inherited methods to worry about
		mi.setRootMethod(mi);
	    }

	    // remember this method's interface, as represented by its
	    // symbol table
	    methods.put(name, mi);
	}
    }
    
    /** look up the method with the given name in this class or one of
     * its superclasses.  return its interface, or return null if a
     * matching method doesn't exist. */
    public MethodInterface lookupMethodOrNull(String name) {
	MethodInterface mi = methods.get(name);
	if (mi == null) {
	    // method not found in this class
	    if (superclass_st == null) {
		// and no superclasses to search, so the method isn't found
		return null;
	    } else {
		// recursively search the superclass
		return superclass_st.lookupMethodOrNull(name);
	    }
	} else {
	    // method found; return its interface
	    return mi;
	}
    }

    /** look up the method with the given name in this class or one of
     * its superclasses.  return the symbol table for the class
     * containing the nearest method definition. */
    public ClassSymbolTable lookupDefiningClass(String name) {
	MethodInterface mi = methods.get(name);
	if (mi == null) {
	    // method not found in this class
	    if (superclass_st == null) {
		// and no superclasses to search, so the method isn't found
		throw new InternalCompilerError(
		    "expecting to find a class defining method \"" + 
		    name + "\"");
	    } else {
		// recursively search the superclass
		return superclass_st.lookupDefiningClass(name);
	    }
	} else {
	    // method found; return the enclosing class's symbol table
	    return this;
	}
    }

    /** declare an instance variable, checking for duplicates */
    public void declareInstanceVariable(ResolvedType type, String name, 
					int line_number)
		throws TypecheckCompilerException {
	// first, declare it, and check for duplicates locally
	declareVariable(name, new InstanceVarInterface(name, type, this),
			line_number);
	// also check that this doesn't shadow any inherited class variables
	checkNoShadowing(name, line_number);
    }

    /** a helper that declares a class variable, checking for
     * duplicates. */
    protected void declareVariable(String name, ClassVarInterface vi,
				   int line_number)
		throws TypecheckCompilerException {
	if (class_variables.containsKey(name)) {
	    // we've already declared a variable with this same name
	    throw new TypecheckCompilerException(
		    "redeclaring class variable named \"" + name + "\"", 
		    line_number);
	} else {
	    // record this new class variable
	    class_variables.put(name, vi);
	}
    }

    /** a helper that checks that a new class variable declaration
     * doesn't shadow any inherited class variable (since MiniJava
     * doesn't support it) */
    protected void checkNoShadowing(String name, int line_number)
		throws TypecheckCompilerException {
	if (superclass_st != null) {
	    ClassVarInterface inherited_vi = 
		superclass_st.lookupClassVariableOrNull(name);
	    if (inherited_vi != null) {
		// we're shadowing an inherited class variable, which
		// isn't supported by MiniJava
		throw new TypecheckCompilerException(
		    "redeclaring inherited class variable named \"" +
		    name + "\"", line_number);
	    } else {
		// there was no inherited class variable, so we're OK
	    }
	} else {
	    // no superclass, so no inherited class variables to worry about
	}
    }

    /** a helper that returns the interface of the class variable
     * named 'name' declared in this class or inherited from a
     * superclass, or returns null if the class variable isn't
     * declared. */
    protected ClassVarInterface lookupClassVariableOrNull(String name) {
	ClassVarInterface vi = class_variables.get(name);
	if (vi == null) {
	    // class variable not declared in this scope.  search up
	    // the inheritance chain.
	    if (superclass_st == null) {
		// there is no superclass; class variable not declared
		return null;
	    } else {
		// search the superclass
		return superclass_st.lookupClassVariableOrNull(name);
	    }
	} else {
	    // found it; return its interface
	    return vi;
	}
    }

    /** Return the type of the variable named 'name' declared in this
     * or some enclosing scope, or throw a TypecheckCompilerException
     * if the variable isn't declared. */
    public VarInterface lookupVariable(String name, int line_number)
		throws TypecheckCompilerException {
	// if we get here, the variable must be a class variable,
	// or be undefined
	ClassVarInterface vi = lookupClassVariableOrNull(name);
	if (vi == null) {
	    // not declared as a class variable, either
	    throw new TypecheckCompilerException(
	        "referencing undefined variable \"" + name + "\"", 
		line_number);
	} else {
	    // found it
	    return vi;
	}
    }


    // Evaluation operations

    /** Initialize all the instance variables in the given object that
     * are declared in this class and its superclasses */
    public void initializeInstanceVariables(ClassValue instance) {
	Iterator<Map.Entry<String,ClassVarInterface>> iter = 
	    class_variables.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,ClassVarInterface> entry = iter.next();
	    ClassVarInterface var_interface = entry.getValue();
	    var_interface.initializeInstanceVariable(instance);
	}
	if (superclass_st != null) {
	    // also initialize all the inherited instance variables
	    superclass_st.initializeInstanceVariables(instance);
	}
    }


    // Lowering (i.e., intermediate-language generation) operations

    // Vtbl-related operations

    /** return the record which is the layout of the virtual function
     * table for this class */
    public ILRecordType vtblRecordType() {
	if (vtbl_type == null) {
	    // compute the class's vtbl record
	    
	    // construct the members for this vtbl
	    List<ILRecordMember> members = new LinkedList<ILRecordMember>();
	    
	    // add fields for all the non-overriding methods of the
	    // class
	    addVtblRecordMembers(members);

	    // create the vtbl record type
	    vtbl_type = new ILRecordType(members);
	}
	// return the vtbl record type
	return vtbl_type;
    }

    /** a recursive helper for the above method */
    protected void addVtblRecordMembers(List<ILRecordMember> members) {
	if (superclass_st != null) {
	    // first, recursively add members for the superclass's
	    // methods (order *does* matter: we need to have the
	    // layout of a superclass's vtbl be a prefix of the layout
	    // of all of its subclasses' vtbls, so that a method
	    // declared in a superclass gets the same position in all
	    // subclasses's vtbls)
	    superclass_st.addVtblRecordMembers(members);
	}

	// add members for each of the locally declared non-overriding
	// methods
	Iterator<Map.Entry<String,MethodInterface>> iter =
	    methods.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,MethodInterface> entry = iter.next();
	    String method_name = entry.getKey();
	    MethodInterface method_interface = entry.getValue();
	    if (method_interface.isRootMethod()) {
		// all functions have a simple lowered type, which ignores
		// their argument and result types
		ILType il_fun_type = new ILPtrType(new ILCodeType());
		members.add(new ILRecordMember(il_fun_type, method_name));
	    } else {
		// don't add a member for an overriding method, since
		// it'll be stored in the same member as its root
		// method
	    }
	}
    }

    /** A helper to compute the value to initialize the class's vtbl with */
    public ILRecordValue vtblRecordValue() {
	List<ILRecordValueMember> members =
	    new LinkedList<ILRecordValueMember>();
	addVtblValueMembers(members, this);
	return new ILRecordValue(members);
    }

    /** A recursive helper for the above method */
    protected void addVtblValueMembers(List<ILRecordValueMember> members,
				       ClassSymbolTable original_class) {
	if (superclass_st != null) {
	    // first, recursively add members for the superclass's
	    // methods (order *does* matter: we need to have the
	    // layout of a superclass's vtbl be a prefix of the layout
	    // of all of its subclasses' vtbls, so that a method
	    // declared in a superclass gets the same position in all
	    // subclasses's vtbls)
	    superclass_st.addVtblValueMembers(members, original_class);
	}

	// add members for each of the locally declared non-overriding
	// methods
	Iterator<Map.Entry<String,MethodInterface>> iter =
	    methods.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,MethodInterface> entry = iter.next();
	    String method_name = entry.getKey();
	    MethodInterface method_interface = entry.getValue();
	    if (method_interface.isRootMethod()) {
		// this is a new entry in the vtbl.  see what method
		// to put there, starting from the original class.
		ClassSymbolTable defining_class = 
		    original_class.lookupDefiningClass(method_name);
		String il_fun_name = 
		    defining_class.getClassName() + "." + method_name;
		ILType il_fun_type = new ILPtrType(new ILCodeType());
		ILValue fun_addr_value =
		    new ILGlobalAddressValue(new ILGlobalVar(il_fun_name));
		members.add(new ILRecordValueMember(il_fun_type,
						    method_name, 
						    fun_addr_value));
	    } else {
		// don't add a member for an overriding method, since
		// we've already stored its value
	    }
	}
    }


    // Instance-related operations

    /** return the record which is the layout of instances of this
     * class */
    public ILRecordType instanceRecordType() {
	if (instance_type == null) {
	    // compute the class's instance record
	    
	    // first create an empty placeholder type

	    // construct the members for this instance
	    List<ILRecordMember> members = new LinkedList<ILRecordMember>();
	    
	    // create the (empty) instance record and record it with
	    // this class (do this early in case of recursive
	    // references to this class's instance layout record while
	    // constructing its members' types)
	    instance_type = new ILRecordType(members);

	    // now fill in the members of this record

	    // add a member for the class's virtual function table
	    members.add(new ILRecordMember(new ILPtrType(vtblRecordType()),
					   "<vtbl>"));

	    // add fields for all the instance variables of the class
	    addInstanceRecordMembers(members);
	}
	// return the instance record type
	return instance_type;
    }

    /** add record members to the argument list for each of the class's
     * instance variables */
    protected void addInstanceRecordMembers(List<ILRecordMember> members) {
	if (superclass_st != null) {
	    // first, recursively add members for the superclass's
	    // instance variables (order *does* matter: we need to
	    // have the layout of a superclass be a prefix of the
	    // layout of all of its subclasses, so that an instance
	    // variable declared in a superclass gets the same
	    // position in all subclasses)
	    superclass_st.addInstanceRecordMembers(members);
	}

	// add members for each of the locally declared instance
	// variables
	Iterator<Map.Entry<String,ClassVarInterface>> iter =
	    class_variables.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,ClassVarInterface> entry = iter.next();
	    String var_name = entry.getKey();
	    ClassVarInterface var_interface = entry.getValue();
	    var_interface.declareInstanceRecordMember(members);
	}
    }

    /** generate code into il_fun_decl to initialize the instance
     * variables of instance_var to their default values */
    public void generateInstanceVariableInits(ILVar instance_var,
					      ILFunDecl il_fun_decl) {
	if (superclass_st != null) {
	    // first, recursively initialize the superclass's instance
	    // variables (order doesn't really matter, but to make the
	    // generated code a little more readable, we'll initialize
	    // top-down)
	    superclass_st.generateInstanceVariableInits(instance_var, 
							il_fun_decl);
	}

	// generate initializing assignments for each of the locally
	// declared instance variables
	Iterator<Map.Entry<String,ClassVarInterface>> iter =
	    class_variables.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,ClassVarInterface> entry = iter.next();
	    ClassVarInterface var_interface = entry.getValue();
	    var_interface.generateInstanceVariableInit(instance_var, 
						       il_fun_decl);
	}
    }
}
