/** The class for the symbol table for the body of a method or of a
 * nested block statement. */

package Typechecker;

import AST.*;
import Helpers.*;
import java.util.*;

public class CodeSymbolTable extends NestedSymbolTable {
    // Instance variables

    /** the interface of local & formal variables declared in this scope */
    protected Map<String,LocalVarInterface> local_variables;


    // Constructor

    /** create the symbol table for a statement list, given the symbol
     * table for the lexically enclosing scope */
    public CodeSymbolTable(SymbolTable enclosing_symbol_table) {
	super(enclosing_symbol_table);
	local_variables = new HashMap<String,LocalVarInterface>();
    }


    // Printing

    /** Return a printable string representation of the symbol table
     * information */
    public String toString(String indent) {
	String s = "CodeSymbolTable{\n";
	String nested_indent = ASTNode.nested_indent(indent);
	Iterator<Map.Entry<String,LocalVarInterface>> iter =
	    local_variables.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,LocalVarInterface> entry = iter.next();
	    String var_name = entry.getKey();
	    LocalVarInterface var_interface = entry.getValue();
	    s += indent + var_name + ": " + var_interface.toString() + "\n";
	}
	s += indent + "}";
	return s;
    }


    // Typechecking operations

    /** declare a local variable, checking for duplicates */
    public LocalVarInterface declareLocalVariable(ResolvedType type, 
						  String name, 
						  int line_number)
		throws TypecheckCompilerException {
	if (local_variables.containsKey(name)) {
	    // we've already declared a local variable with this same name
	    throw new TypecheckCompilerException(
		    "redeclaring local variable named \"" + name + "\"", 
		    line_number);
	} else {
	    // record this new local variable
	    LocalVarInterface vi = new LocalVarInterface(name, type);
	    local_variables.put(name, vi);
	    return vi;
	}
    }

    /** Return the interface for the variable named 'name' declared in
     * this or some enclosing scope, or throw a TypecheckCompilerException if
     * the variable isn't declared. */
    public VarInterface lookupVariable(String name, int line_number)
		throws TypecheckCompilerException {
	LocalVarInterface vi = local_variables.get(name);
	if (vi == null) {
	    // variable not declared in this scope; search the
	    // lexically enclosing scope
	    return enclosing_symbol_table.lookupVariable(name, line_number);
	} else {
	    // found it; return its type
	    return vi;
	}
    }
}
