/** The class for the symbol table for the global scope, which stores
 * information about classes. */

package Typechecker;

import AST.*;
import Helpers.*;
import java.util.*;

public class GlobalSymbolTable extends SymbolTable {
    // Instance variables

    /** A map from class names to the symbol tables for each of the
     * top-level classes in the program */
    protected Map<String,ClassSymbolTable> classes;


    // Constructor

    public GlobalSymbolTable() {
	classes = new HashMap<String,ClassSymbolTable>();
    }


    // Printing

    /** Return a printable string representation of the symbol table
     * information */
    public String toString(String indent) {
	String s = "GlobalSymbolTable{\n";
	String nested_indent = ASTNode.nested_indent(indent);
	Iterator<Map.Entry<String,ClassSymbolTable>> iter =
	    classes.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry<String,ClassSymbolTable> entry = iter.next();
	    String class_name = entry.getKey();
	    ClassSymbolTable class_interface = entry.getValue();
	    s += nested_indent + class_name + ": " + 
		class_interface.toString(nested_indent) + "\n";
	}
	s += indent + "}";
	return s;
    }


    // Typechecking operations

    /** declare a new class in the global scope.  do
     * duplicate-class-declaration error-checking. */
    public void declareClass(String class_name, ClassSymbolTable st, 
			     int line_number)
		throws TypecheckCompilerException {
	if (classes.containsKey(class_name)) {
	    // we've already declared a class with this same name
	    throw new TypecheckCompilerException(
		"redeclaring class named \"" + class_name + "\"",
		line_number);
	} else {
	    // add the class's symbol table to the classes map
	    classes.put(class_name, st);
	}
    }

    /** Return the symbol table for the class named 'class_name', or
     * throw a TypecheckCompilerException if the class isn't declared. */
    public ClassSymbolTable lookupClass(String name, int line_number) 
		throws TypecheckCompilerException {
	ClassSymbolTable st = classes.get(name);
	if (st == null) {
	    // class not found
	    throw new TypecheckCompilerException(
		    "referencing undefined class \"" + name + "\"",
		    line_number);
	} else {
	    // class found
	    return st;
	}
    }

    /** Return the variable interface for the variable named 'name'
     * declared in this or some enclosing scope, or throw a
     * TypecheckCompilerException if the variable isn't declared. */
    public VarInterface lookupVariable(String name, int line_number)
		throws TypecheckCompilerException {
	// variable not declared
	throw new TypecheckCompilerException(
	    "referencing undefined variable \"" + name + "\"",  line_number);
    }
}
