/** The class for information about instance variables. */

package Typechecker;

import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class InstanceVarInterface extends ClassVarInterface {
    // Constructor
    public InstanceVarInterface(String name, ResolvedType type, 
				ClassSymbolTable cst) {
	super(name, type, cst);
    }


    // Evaluation operations

    /** Initialize this class variable in the given instance, if this
     * is an instance variable */
    public void initializeInstanceVariable(ClassValue instance) {
	instance.initializeInstanceVariable(name, type);
    }

    /** Return the current contents of this variable */
    public Value lookupVariable(CodeEnvironment env, int line_number) 
		throws EvalCompilerException {
	Value receiver;
	try {
	    // get the implicit reciever whose instance variable is
	    // being accessed
	    receiver = env.lookupLocalVariable("this", line_number);
	} catch (EvalCompilerException e) {
	    // not expecting this to fail
	    throw new InternalCompilerError(e.getMessage(), line_number);
	}
	// read the contents of the receiver's instance variable
	return receiver.lookupInstanceVariable(name, line_number);
    }

    /** Update the current contents of this variable */
    public void assignVariable(CodeEnvironment env, Value value,
			       int line_number) {
	Value receiver;
	try {
	    // get the implicit reciever whose instance variable is
	    // being accessed
	    receiver = env.lookupLocalVariable("this", line_number);
	} catch (EvalCompilerException e) {
	    // not expecting this to fail
	    throw new InternalCompilerError(e.getMessage(), line_number);
	}
	// assign to the contents of the receiver's instance variable
	receiver.assignInstanceVariable(name, value, line_number);
    }


    // Lowering (i.e., intermediate-language generation) operations

    /** generate intermediate-language code to assign to this
     * variable */
    public void generateAssignment(ILExpr rhs_expr, ILFunDecl il_fun_decl) {
	// get the var for the receiver
	ILVar receiver_var = il_fun_decl.lookupThis();

	// (we can assume that the receiver isn't null, since we're
	// inside a method defined on the receiver)

	// get the record member for this instance variable
	ILType class_type = class_st.instanceRecordType();
	ILRecordMember var_member = class_type.getRecordMember(name);

	// generate a field store
	ILAssignableExpr var_field =
	    new ILFieldAccessExpr(new ILVarExpr(receiver_var), 
				  class_type, var_member);
	il_fun_decl.addStmt(new ILAssignStmt(var_field, rhs_expr));
    }

    /** generate intermediate-language code to read from this
     * variable, returning an IL variable holding the result */
    public ILExpr generateRead(ILFunDecl il_fun_decl) {
	// get the var for the receiver
	ILVar receiver_var = il_fun_decl.lookupThis();

	// (we can assume that the receiver isn't null, since we're
	// inside a method defined on the receiver.)

	// get the record member for this instance variable
	ILType class_type = class_st.instanceRecordType();
	ILRecordMember var_member = class_type.getRecordMember(name);

	// construct code to load the contents of this field
	ILExpr var_field =
	    new ILFieldAccessExpr(new ILVarExpr(receiver_var), 
				  class_type, var_member);

	// since other expressions (e.g. calls) might modify this
	// field, we need to execute this expression in the right
	// order, so generate a statement that loads this field into a
	// temp, and return that
	ILVar result_var = il_fun_decl.newTempVar(var_member.getType());
	il_fun_decl.addStmt(new ILAssignStmt(new ILVarExpr(result_var),
					     var_field));
	return new ILVarExpr(result_var);
    }


    /** if this is an instance variable, generate code into
     * il_fun_decl to initialize this instance variable of
     * instance_var to its default value */
    public void generateInstanceVariableInit(ILVar instance_var,
					     ILFunDecl il_fun_decl) {
	// get the record member for this instance variable
	ILType class_type = class_st.instanceRecordType();
	ILRecordMember var_member = class_type.getRecordMember(name);

	// compute the default initial value for this instance variable based
	// on its type
	ILExpr init_expr = type.defaultILExpr();

	// generate a field store
	ILAssignableExpr var_field =
	    new ILFieldAccessExpr(new ILVarExpr(instance_var), 
				  class_type, var_member);
	il_fun_decl.addStmt(new ILAssignStmt(var_field, init_expr));
    }

    /** if this is an instance variable, add a declaration for this
     * variable's member into the members list of the class instance
     * record */
    public void declareInstanceRecordMember(List<ILRecordMember> members) {
	// add a member to the instance record for this instance variable
	members.add(new ILRecordMember(type.lower(), name));
    }
}
