/** The class for (resolved) class types. */

package Typechecker;

import Evaluator.*;
import IL.*;
import Helpers.*;
import java.util.*;

public class ResolvedClassType extends ResolvedReferenceType {
    // the symbol table for the class that this type is for
    protected ClassSymbolTable class_interface;

    // a cache of the lowered version of this type, computed during
    // lowering
    protected ILPtrType lowered_type;


    // Constructors
    public ResolvedClassType(ClassSymbolTable ci) {
	class_interface = ci;
    }
    public ResolvedClassType(String class_name, 
			     SymbolTable st,
			     int line_number)
		throws TypecheckCompilerException { 
	// call the other constructor, after looking up the class's
	// interface from its name
	this(st.lookupClass(class_name, line_number));
    }


    // Printing operations
    public String toString() { 
	return class_interface.getClassName(); 
    }


    // Testers

    /** return whether this type is a class type */
    public boolean isClassType() {
	return true;
    }


    // Accessors

    /** Return the interface to the named class, if a class type */
    public ClassSymbolTable getClassInterface() {
	return class_interface;
    }


    // Type comparison operations

    /** Return whether two types should be considered the same type */
    public boolean isSameType(ResolvedType other) {
	if (other instanceof ResolvedClassType) {
	    ResolvedClassType other_class_type = (ResolvedClassType)other;
	    // two class types are the same if they're from the same class
	    return getClassInterface() == other_class_type.getClassInterface();
	} else {
	    // not equal types
	    return false;
	}
    }


    /** Return whether the receiver should be considered a subtype of
     * the argument */
    public boolean isSubtype(ResolvedType other) {
	if (other instanceof ResolvedClassType) {
	    ResolvedClassType other_class_type = (ResolvedClassType)other;
	    // one class type is a subtype of the other if they're the
	    // same or the first inherits (directly or indirectly)
	    // from the other
	    ClassSymbolTable ci = getClassInterface();
	    ClassSymbolTable other_ci = other_class_type.getClassInterface();
	    return ci == other_ci || ci.inheritsFrom(other_ci);
	} else {
	    // not subtypes
	    return false;
	}
    }
    

    // Typechecking operations

    /** look up the method with the given name in the receiver type.
     * return its interface, or throw an exception if a matching
     * method doesn't exist. */
    public MethodInterface lookupMethod(String name, int line_number)
		throws TypecheckCompilerException {
	// the receiver is a class type, so do the look-up in the
	// class's interface
	MethodInterface mi = class_interface.lookupMethodOrNull(name);
	if (mi == null) {
	    // a matching method wasn't found
	    throw new TypecheckCompilerException(
		    "calling undefined method \"" + name + "\"",
		    line_number);
	} else {
	    // found it
	    return mi;
	}
    }


    // Evaluation operations

    /** Return the default initial value for a variable of this type */
    public Value defaultValue() {
	// null is the default value
	return new NullValue();
    }


    // Lowering operations

    /** Return the lowered version of this type */
    public ILType lower() {
	// an instance of a class is represented by a pointer to a
	// record derived from the class's instance variables
	if (lowered_type == null) {
	    lowered_type = new ILPtrType(class_interface.instanceRecordType());
	}
	return lowered_type;
    }

    /** Return an IL expression representing the default value of this
     * type */
    public ILExpr defaultILExpr() {
	// null is the default value, which is represented in the
	// lowered code by the integer 0
	return new ILIntConstantExpr(0);
    }

    /** Return an IL value for initializing a global declaration
     * representing the default value of this type */
    public ILValue defaultILValue() {
	// null is the default value, which is represented in the
	// lowered code by the integer 0
	return new ILIntValue(0);
    }
}
