/** This class represents types that have been checked for correctness
 * as legal types. */

package Typechecker;

import Evaluator.*;
import IL.*;
import Helpers.*;

public abstract class ResolvedType {
    // Printing operations
    public abstract String toString();


    // Testers

    /** return whether this type is an int type */
    public boolean isIntType() {
	// by default, not
	return false;
    }

    /** return whether this type is an int type */
    public boolean isBooleanType() {
	// by default, not
	return false;
    }

    /** return whether this type is a class type */
    public boolean isClassType() {
	// by default, not
	return false;
    }


    // Accessors

    /** Return the interface to the named class, if a class type */
    public ClassSymbolTable getClassInterface() {
	throw new InternalCompilerError("not a class type");
    }
    

    // Type comparison operations

    /** Return whether two types should be considered the same type */
    public abstract boolean isSameType(ResolvedType other);

    /** Return whether the receiver should be considered a subtype of
     * the argument */
    public boolean isSubtype(ResolvedType other) {
	// by default, one resolved type is a subtype of another only
	// if they're the same type.  overridden where this default
	// doesn't hold, e.g. for class types.
	return isSameType(other);
    }

    /** Return whether the receiver can be assigned to or used in a
     * context expecting the argument.  this is true if the receiver
     * is a subtype of the argument, or if the receiver can be
     * converted to the argument type (in which case an implicit
     * conversion should be inserted). */
    public boolean isAssignableTo(ResolvedType other) {
	// for now, no converstions allowed; just subtyping
	return isSubtype(other);
    }


    // Typechecking operations

    /** Verify that the receiver can be assigned to or used in a
     * context expecting the argument, throwing an exception if not.
     * this is true if the receiver is a subtype of the argument, or
     * if the receiver can be converted to the argument type (in which
     * case an implicit conversion should be inserted) */
    public void checkIsAssignableTo(ResolvedType other,
				    String this_name,
				    String other_name, 
				    int line_number) 
		throws TypecheckCompilerException {
	if (! isAssignableTo(other)) {
	    throw new TypecheckCompilerException(
		this_name + " is not assignable to " + other_name, 
		line_number);
	}
    }

    /** Verify that the receiver is an int type, throwing an
     * exception if not */
    public void checkIsIntType(String this_name, int line_number) 
		throws TypecheckCompilerException {
	if (! isIntType()) {
	    throw new TypecheckCompilerException(
		this_name + " is not an int", line_number);
	}
    }

    /** Verify that the receiver is a boolean type, throwing an
     * exception if not */
    public void checkIsBooleanType(String this_name, int line_number) 
		throws TypecheckCompilerException {
	if (! isBooleanType()) {
	    throw new TypecheckCompilerException(
		this_name + " is not a boolean", line_number);
	}
    }

    /** Verify that the receiver is a class type, throwing an
     * exception if not */
    public void checkIsClassType(String this_name, int line_number) 
		throws TypecheckCompilerException {
	if (! isClassType()) {
	    throw new TypecheckCompilerException(
		this_name + " is not a class instance", line_number);
	}
    }
    
    /** look up the method with the given name in the receiver type.
     * return its interface, or throw an exception if a matching
     * method doesn't exist. */
    public MethodInterface lookupMethod(String name, int line_number)
		throws TypecheckCompilerException {
	// by default, it's not a class type, so it can't be sent
	// messages
	throw new TypecheckCompilerException(
	    "cannot invoke a method on a non-class receiver", line_number);
    }


    // Evaluation operations

    /** Return the default initial value for a variable of this
     * type */
    public abstract Value defaultValue();


    // Lowering operations

    /** Return the lowered version of this type */
    public abstract ILType lower();

    /** Return an IL expression representing the default value of this
     * type */
    public abstract ILExpr defaultILExpr();

    /** Return an IL value for initializing a global declaration
     * representing the default value of this type */
    public abstract ILValue defaultILValue();
}
